"""
Callbacks to add additional functionality on to plots.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: J. S. Oishi <jsoishi@astro.berkeley.edu>
Affiliation: UC Berkeley
Author: Stephen Skory <sskory@physics.ucsd.edu>
Affiliation: UC San Diego
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk, JS Oishi, Stephen Skory.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.raven import *
from PlotTypes import _get_bounds

import _MPL
import copy
callback_registry = {}

class PlotCallback(object):
    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            callback_registry[name] = cls

    def __init__(self, *args, **kwargs):
        pass

    def convert_to_pixels(self, plot, coord, offset = True):
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        l, b, width, height = _get_bounds(plot._axes.bbox)
        dx = width / (x1-x0)
        dy = height / (y1-y0)
        return ((coord[0] - int(offset)*x0)*dx,
                (coord[1] - int(offset)*y0)*dy)

class VelocityCallback(PlotCallback):
    _type_name = "velocity"
    def __init__(self, factor=16):
        """
        Adds a 'quiver' plot of velocity to the plot, skipping all but
        every *factor* datapoint
        """
        PlotCallback.__init__(self)
        self.factor = factor

    def __call__(self, plot):
        # Instantiation of these is cheap
        if plot._type_name == "CuttingPlane":
            qcb = CuttingQuiverCallback("CuttingPlaneVelocityX",
                                        "CuttingPlaneVelocityY",
                                        self.factor)
        else:
            xv = "%s-velocity" % (lagos.x_names[plot.data.axis])
            yv = "%s-velocity" % (lagos.y_names[plot.data.axis])
            qcb = QuiverCallback(xv, yv, self.factor)
        return qcb(plot)

class QuiverCallback(PlotCallback):
    _type_name = "quiver"
    def __init__(self, field_x, field_y, factor):
        """
        Adds a 'quiver' plot to any plot, using the *field_x* and *field_y*
        from the associated data, skipping every *factor* datapoints.
        """
        PlotCallback.__init__(self)
        self.field_x = field_x
        self.field_y = field_y
        self.bv_x = self.bv_y = 0
        self.factor = factor

    def __call__(self, plot):
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        xx0, xx1 = plot._axes.get_xlim()
        yy0, yy1 = plot._axes.get_ylim()
        plot._axes.hold(True)
        nx = plot.image._A.shape[0] / self.factor
        ny = plot.image._A.shape[1] / self.factor
        pixX = _MPL.Pixelize(plot.data['px'],
                             plot.data['py'],
                             plot.data['pdx'],
                             plot.data['pdy'],
                             plot.data[self.field_x] - self.bv_x,
                             int(nx), int(ny),
                           (x0, x1, y0, y1),).transpose()
        pixY = _MPL.Pixelize(plot.data['px'],
                             plot.data['py'],
                             plot.data['pdx'],
                             plot.data['pdy'],
                             plot.data[self.field_y] - self.bv_y,
                             int(nx), int(ny),
                           (x0, x1, y0, y1),).transpose()
        X = na.mgrid[0:plot.image._A.shape[0]-1:nx*1j]# + 0.5*factor
        Y = na.mgrid[0:plot.image._A.shape[1]-1:ny*1j]# + 0.5*factor
        plot._axes.quiver(X,Y, pixX, pixY)
        plot._axes.set_xlim(xx0,xx1)
        plot._axes.set_ylim(yy0,yy1)
        plot._axes.hold(False)

class ContourCallback(PlotCallback):
    _type_name = "contour"
    def __init__(self, field, ncont=5, factor=4, take_log=False, clim=None,
                 plot_args = None):
        """
        Add contours in *field* to the plot.  *ncont* governs the number of
        contours generated, *factor* governs the number of points used in the
        interpolation, *take_log* governs how it is contoured and *clim* gives
        the (upper, lower) limits for contouring.
        """
        PlotCallback.__init__(self)
        self.ncont = ncont
        self.field = field
        self.factor = factor
        self.take_log = take_log
        try:
            import delaunay as de
            self.de = de
        except ImportError:
            mylog.warning("Callback failed; no delaunay module")
            self.__call__ = lambda a: None
        if self.take_log and clim is not None: clim = (na.log10(clim[0]), na.log10(clim[1]))
        if clim is not None: self.ncont = na.linspace(clim[0], clim[1], ncont)
        if plot_args is None: plot_args = {'colors':'k'}
        self.plot_args = plot_args

    def __call__(self, plot):
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        print x0, x1, y0, y1
        xx0, xx1 = plot._axes.get_xlim()
        yy0, yy1 = plot._axes.get_ylim()
        plot._axes.hold(True)
        numPoints_x = plot.image._A.shape[0]
        numPoints_y = plot.image._A.shape[1]
        dy = plot.image._A.shape[0] / (x1-x0)
        dx = plot.image._A.shape[1] / (y1-y0)
        #dcollins Jan 11 2009.  Improved to allow for periodic shifts in the plot.
        #Now makes a copy of the position fields "px" and "py" and adds the
        #appropriate shift to the coppied field.  

        #set the cumulative arrays for the periodic shifting.
        AllX = na.zeros(plot.data["px"].size, dtype='bool')
        AllY = na.zeros(plot.data["py"].size, dtype='bool')
        XShifted = plot.data["px"].copy()
        YShifted = plot.data["py"].copy()
        dom_x, dom_y = plot._period
        for shift in na.mgrid[-1:1:3j]:
            xlim = ((plot.data["px"] + shift*dom_x >= x0)
                 &  (plot.data["px"] + shift*dom_x <= x1))
            ylim = ((plot.data["py"] + shift*dom_y >= y0)
                 &  (plot.data["py"] + shift*dom_y <= y1))
            XShifted[xlim] += shift * dom_x
            YShifted[ylim] += shift * dom_y
            AllX |= xlim
            AllY |= ylim
        wI = (AllX & AllY)
        xi, yi = na.mgrid[0:numPoints_x:numPoints_x/(self.factor*1j),\
                          0:numPoints_y:numPoints_y/(self.factor*1j)]
        x = (XShifted[wI]-x0)*dx 
        y = (YShifted[wI]-y0)*dy
        z = plot.data[self.field][wI]
        if self.take_log: z=na.log10(z)
        zi = self.de.Triangulation(x,y).nn_interpolator(z)(xi,yi)
        print z.min(), z.max(), na.nanmin(z), na.nanmax(z)
        print zi.min(), zi.max(), na.nanmin(zi), na.nanmax(zi)
        plot._axes.contour(xi,yi,zi,self.ncont, **self.plot_args)
        plot._axes.set_xlim(xx0,xx1)
        plot._axes.set_ylim(yy0,yy1)
        plot._axes.hold(False)

class GridBoundaryCallback(PlotCallback):
    _type_name = "grids"
    def __init__(self, alpha=1.0, min_pix=1, annotate=False, periodic=True):
        """
        Adds grid boundaries to a plot, optionally with *alpha*-blending.
        Cuttoff for display is at *min_pix* wide.
        *annotate* puts the grid id in the corner of the grid.  (Not so great in projections...)
        """
        PlotCallback.__init__(self)
        self.alpha = alpha
        self.min_pix = min_pix
        self.annotate = annotate # put grid numbers in the corner.
        self.periodic = periodic

    def __call__(self, plot):
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        xx0, xx1 = plot._axes.get_xlim()
        yy0, yy1 = plot._axes.get_ylim()
        dx = (xx1-xx0)/(x1-x0)
        dy = (yy1-yy0)/(y1-y0)
        px_index = lagos.x_dict[plot.data.axis]
        py_index = lagos.y_dict[plot.data.axis]
        dom = plot.data.pf["DomainRightEdge"] - plot.data.pf["DomainLeftEdge"]
        if self.periodic:
            pxs, pys = na.mgrid[-1:1:3j,-1:1:3j]
        else:
            pxs, pys = na.mgrid[0:0:1j,0:0:1j]
        GLE = plot.data.grid_left_edge
        GRE = plot.data.grid_right_edge
        for px_off, py_off in zip(pxs.ravel(), pys.ravel()):
            pxo = px_off * dom[px_index]
            pyo = py_off * dom[py_index]
            left_edge_px = (GLE[:,px_index]+pxo-x0)*dx
            left_edge_py = (GLE[:,py_index]+pyo-y0)*dy
            right_edge_px = (GRE[:,px_index]+pxo-x0)*dx
            right_edge_py = (GRE[:,py_index]+pyo-y0)*dy
            verts = na.array(
                    [(left_edge_px, left_edge_px, right_edge_px, right_edge_px),
                     (left_edge_py, right_edge_py, right_edge_py, left_edge_py)])
            visible =  ( right_edge_px - left_edge_px > self.min_pix ) & \
                       ( right_edge_px - left_edge_px > self.min_pix )
            verts=verts.transpose()[visible,:,:]
            if verts.size == 0: continue
            edgecolors = (0.0,0.0,0.0,self.alpha)
            grid_collection = matplotlib.collections.PolyCollection(
                    verts, facecolors="none",
                           edgecolors=edgecolors)
            plot._axes.hold(True)
            plot._axes.add_collection(grid_collection)
            if self.annotate:
                ids = [g.id for g in plot.data._grids]
                for n in range(len(left_edge_px)):
                    plot._axes.text(left_edge_px[n]+2,left_edge_py[n]+2,ids[n])
            plot._axes.hold(False)

class LabelCallback(PlotCallback):
    _type_name = "axis_label"
    def __init__(self, label):
        """
        This adds a label to the plot.
        """
        PlotCallback.__init__(self)
        self.label = label

    def __call__(self, plot):
        plot._figure.subplots_adjust(hspace=0, wspace=0,
                                     bottom=0.1, top=0.9,
                                     left=0.0, right=1.0)
        plot._axes.set_xlabel(self.label)
        plot._axes.set_ylabel(self.label)

def get_smallest_appropriate_unit(v, pf):
    max_nu = 1e30
    good_u = None
    for unit in ['mpc','kpc','pc','au','rsun','cm']:
        vv = v*pf[unit]
        if vv < max_nu and vv > 1.0:
            good_u = unit
            max_nu = v*pf[unit]
    return good_u

class UnitBoundaryCallback(PlotCallback):
    _type_name = "units"
    def __init__(self, unit = "au", factor=4, text_annotate=True, text_which=-2):
        """
        Add on a plot indicating where *factor*s of *unit* are shown.
        Optionally *text_annotate* on the *text_which*-indexed box on display.
        """
        PlotCallback.__init__(self)
        self.unit = unit
        self.factor = factor
        self.text_annotate = text_annotate
        self.text_which = -2

    def __call__(self, plot):
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        l, b, width, height = _get_bounds(plot._axes.bbox)
        xi = lagos.x_dict[plot.data.axis]
        yi = lagos.y_dict[plot.data.axis]
        dx = plot.image._A.shape[0] / (x1-x0)
        dy = plot.image._A.shape[1] / (y1-y0)
        center = plot.data.center
        min_dx = plot.data['pdx'].min()
        max_dx = plot.data['pdx'].max()
        w_min_x = 250.0 * min_dx
        w_max_x = 1.0 / self.factor
        min_exp_x = na.ceil(na.log10(w_min_x*plot.data.pf[self.unit])
                           /na.log10(self.factor))
        max_exp_x = na.floor(na.log10(w_max_x*plot.data.pf[self.unit])
                            /na.log10(self.factor))
        n_x = max_exp_x - min_exp_x + 1
        widths = na.logspace(min_exp_x, max_exp_x, num = n_x, base=self.factor)
        widths /= plot.data.pf[self.unit]
        left_edge_px = (center[xi] - widths/2.0 - x0)*dx
        left_edge_py = (center[yi] - widths/2.0 - y0)*dy
        right_edge_px = (center[xi] + widths/2.0 - x0)*dx
        right_edge_py = (center[yi] + widths/2.0 - y0)*dy
        verts = na.array(
                [(left_edge_px, left_edge_px, right_edge_px, right_edge_px),
                 (left_edge_py, right_edge_py, right_edge_py, left_edge_py)])
        visible =  ( right_edge_px - left_edge_px > 25 ) & \
                   ( right_edge_px - left_edge_px > 25 ) & \
                   ( (right_edge_px < width) & (left_edge_px > 0) ) & \
                   ( (right_edge_py < height) & (left_edge_py > 0) )
        verts=verts.transpose()[visible,:,:]
        grid_collection = matplotlib.collections.PolyCollection(
                verts, facecolors="none",
                       edgecolors=(0.0,0.0,0.0,1.0),
                       linewidths=2.5)
        plot._axes.hold(True)
        plot._axes.add_collection(grid_collection)
        if self.text_annotate:
            ti = max(self.text_which, -1*len(widths[visible]))
            if ti < len(widths[visible]): 
                w = widths[visible][ti]
                good_u = get_smallest_appropriate_unit(w, plot.data.pf)
                w *= plot.data.pf[good_u]
                plot._axes.annotate("%0.3e %s" % (w,good_u), verts[ti,1,:]+5)
        plot._axes.hold(False)

class LinePlotCallback(PlotCallback):
    _type_name = "line"
    def __init__(self, x, y, plot_args = None):
        """
        Over plot *x* and *y* with *plot_args* fed into the plot.
        """
        PlotCallback.__init__(self)
        self.x = x
        self.y = y
        if plot_args is None: plot_args = {}
        self.plot_args = plot_args

    def __call__(self, plot):
        plot._axes.hold(True)
        plot._axes.plot(self.x, self.y, **self.plot_args)
        plot._axes.hold(False)

class ImageLineCallback(LinePlotCallback):
    _type_name = "image_line"

    def __init__(self, p1, p2, plot_args = None):
        """
        Plot from *p1* to *p2* (image plane coordinates)
        with *plot_args* fed into the plot.
        """
        PlotCallback.__init__(self)
        self.p1 = p1
        self.p2 = p2
        if plot_args is None: plot_args = {}
        self.plot_args = plot_args
        self._ids = []

    def __call__(self, plot):
        # We manually clear out any previous calls to this callback:
        plot._axes.lines = [l for l in plot._axes.lines if id(l) not in self._ids]
        p1 = self.convert_to_pixels(plot, self.p1)
        p2 = self.convert_to_pixels(plot, self.p2)
        px, py = (p1[0], p2[0]), (p1[1], p2[1])

        # Save state
        xx0, xx1 = plot._axes.get_xlim()
        yy0, yy1 = plot._axes.get_ylim()
        plot._axes.hold(True)
        ii = plot._axes.plot(px, py, **self.plot_args)
        self._ids.append(id(ii[0]))
        # Reset state
        plot._axes.set_xlim(xx0,xx1)
        plot._axes.set_ylim(yy0,yy1)
        plot._axes.hold(False)

class CuttingQuiverCallback(PlotCallback):
    _type_name = "cquiver"
    def __init__(self, field_x, field_y, factor):
        """
        Get a quiver plot on top of a cutting plane, using *field_x* and
        *field_y*, skipping every *factor* datapoint in the discretization.
        """
        PlotCallback.__init__(self)
        self.field_x = field_x
        self.field_y = field_y
        self.factor = factor

    def __call__(self, plot):
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        xx0, xx1 = plot._axes.get_xlim()
        yy0, yy1 = plot._axes.get_ylim()
        plot._axes.hold(True)
        nx = plot.image._A.shape[0] / self.factor
        ny = plot.image._A.shape[1] / self.factor
        indices = na.argsort(plot.data['dx'])[::-1]
        pixX = _MPL.CPixelize( plot.data['x'], plot.data['y'], plot.data['z'],
                               plot.data['px'], plot.data['py'],
                               plot.data['pdx'], plot.data['pdy'], plot.data['pdz'],
                               plot.data.center, plot.data._inv_mat, indices,
                               plot.data[self.field_x],
                               int(nx), int(ny),
                               (x0, x1, y0, y1),).transpose()
        pixY = _MPL.CPixelize( plot.data['x'], plot.data['y'], plot.data['z'],
                               plot.data['px'], plot.data['py'],
                               plot.data['pdx'], plot.data['pdy'], plot.data['pdz'],
                               plot.data.center, plot.data._inv_mat, indices,
                               plot.data[self.field_y],
                               int(nx), int(ny),
                               (x0, x1, y0, y1),).transpose()
        X = na.mgrid[0:plot.image._A.shape[0]-1:nx*1j]# + 0.5*factor
        Y = na.mgrid[0:plot.image._A.shape[1]-1:ny*1j]# + 0.5*factor
        plot._axes.quiver(X,Y, pixX, pixY)
        plot._axes.set_xlim(xx0,xx1)
        plot._axes.set_ylim(yy0,yy1)
        plot._axes.hold(False)

class ClumpContourCallback(PlotCallback):
    _type_name = "clumps"
    def __init__(self, clumps, plot_args = None):
        """
        Take a list of *clumps* and plot them as a set of contours.
        """
        self.clumps = clumps
        if plot_args is None: plot_args = {}
        self.plot_args = plot_args

    def __call__(self, plot):
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        xx0, xx1 = plot._axes.get_xlim()
        yy0, yy1 = plot._axes.get_ylim()
        plot._axes.hold(True)

        px_index = lagos.x_dict[plot.data.axis]
        py_index = lagos.y_dict[plot.data.axis]

        xf = lagos.axis_names[px_index]
        yf = lagos.axis_names[py_index]

        DomainRight = plot.data.pf["DomainRightEdge"]
        DomainLeft = plot.data.pf["DomainLeftEdge"]
        DomainWidth = DomainRight - DomainLeft
        
        nx, ny = plot.image._A.shape
        buff = na.zeros((nx,ny),dtype='float64')
        for i,clump in enumerate(reversed(self.clumps)):
            mylog.debug("Pixelizing contour %s", i)


            xf_copy = copy.copy(clump[xf])
            yf_copy = copy.copy(clump[yf])

            #Shift zones that belong shifted, both directions in X and Y.
            shifted = na.logical_and( xf_copy + DomainWidth[px_index] >= DomainRight[px_index],
                                      xf_copy + DomainWidth[px_index]<= x1 )
            xf_copy[na.where(shifted)] += DomainWidth[px_index]
            
            shifted = na.logical_and( xf_copy - DomainWidth[px_index] <= DomainLeft[px_index],
                                      xf_copy - DomainWidth[px_index] >= x0 )
            xf_copy[na.where(shifted)] -= DomainWidth[px_index]
            
            shifted = na.logical_and( yf_copy + DomainWidth[py_index] >= DomainRight[py_index],
                                      yf_copy + DomainWidth[py_index] <= y1 )
            yf_copy[na.where(shifted)] += DomainWidth[py_index]
            
            shifted = na.logical_and( yf_copy - DomainWidth[py_index] <= DomainLeft[py_index],
                                      yf_copy - DomainWidth[py_index] >= y0 )
            yf_copy[na.where(shifted)] -= DomainWidth[py_index]
            
            temp = _MPL.Pixelize(xf_copy, yf_copy, 
                                 clump['dx']/2.0,
                                 clump['dy']/2.0,
                                 clump['dx']*0.0+i+1, # inits inside Pixelize
                                 int(nx), int(ny),
                             (x0, x1, y0, y1), 0).transpose()
            buff = na.maximum(temp, buff)
        self.rv = plot._axes.contour(buff, len(self.clumps)+1,
                                     **self.plot_args)
        plot._axes.hold(False)

class ArrowCallback(PlotCallback):
    _type_name = "arrow"
    def __init__(self, pos, code_size, plot_args = None):
        """
        This adds an arrow pointing at *pos* with size *code_size* in code
        units.  *plot_args* is a dict fed to matplotlib with arrow properties.
        """
        self.pos = pos
        self.code_size = code_size
        if plot_args is None: plot_args = {}
        self.plot_args = plot_args

    def __call__(self, plot):
        from matplotlib.patches import Arrow
        # Now convert the pixels to code information
        x, y = self.convert_to_pixels(plot, self.pos)
        dx, dy = self.convert_to_pixels(plot, self.code_size, False)
        arrow = Arrow(x, y, dx, dy, **self.plot_args)
        plot._axes.add_patch(arrow)

class PointAnnotateCallback(PlotCallback):
    _type_name = "point"
    def __init__(self, pos, text, text_args = None):
        """
        This adds *text* at position *pos*, where *pos* is in code-space.
        *text_args* is a dict fed to the text placement code.
        """
        self.pos = pos
        self.text = text
        self.text_args = text_args

    def __call__(self, plot):
        x,y = self.convert_to_pixels(plot, self.pos)
        plot._axes.text(x, y, self.text, **self.text_args)

class MarkerAnnotateCallback(PlotCallback):
    _type_name = "marker"
    def __init__(self, pos, marker='x', plot_args=None):
        """
        Adds text *marker* at *pos* in code-arguments.  *plot_args* is a dict
        that will be forwarded to the plot command.
        """
        self.pos = pos
        self.marker = marker
        if plot_args is None: plot_args = {}
        self.plot_args = plot_args

    def __call__(self, plot):
        if len(self.pos) == 3:
            pos = (self.pos[lagos.x_dict[plot.data.axis]],
                   self.pos[lagos.y_dict[plot.data.axis]])
        else: pos = self.pos
        x,y = self.convert_to_pixels(plot, pos)
        print x, y
        plot._axes.hold(True)
        plot._axes.plot((x,),(y,),self.marker, **self.plot_args)
        plot._axes.hold(False)

class SphereCallback(PlotCallback):
    _type_name = "sphere"
    def __init__(self, center, radius, circle_args = None,
                 text = None, text_args = None):
        """
        A sphere centered at *center* in code units with radius *radius* in
        code units will be created, with optional *circle_args*, *text*, and
        *text_args*.
        """
        self.center = center
        self.radius = radius
        if circle_args is None: circle_args = {}
        if 'fill' not in circle_args: circle_args['fill'] = False
        self.circle_args = circle_args
        self.text = text
        self.text_args = text_args

    def __call__(self, plot):
        from matplotlib.patches import Circle
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        l, b, width, height = _get_bounds(plot._axes.bbox)
        xi = lagos.x_dict[plot.data.axis]
        yi = lagos.y_dict[plot.data.axis]
        dx = plot.image._A.shape[0] / (x1-x0)
        dy = plot.image._A.shape[1] / (y1-y0)
        radius = self.radius * dx
        center_x = (self.center[xi] - x0)*dx
        center_y = (self.center[yi] - y0)*dy
        # origin = lower?  not sure why center_y and center_x are reversed
        cir = Circle((center_x, center_y), radius, **self.circle_args)
        plot._axes.add_patch(cir)
        if self.text is not None:
            plot._axes.text(center_x, center_y, "%s" % halo.id,
                            **self.text_args)

class HopCircleCallback(PlotCallback):
    _type_name = "hop_circles"
    def __init__(self, hop_output, max_number=None,
                 annotate=False, min_size=20, max_size=10000000,
                 font_size=8, print_halo_size=False,
                 print_halo_mass=False, width=None):
        """
        Accepts a :class:`yt.lagos.HopList` *hop_output* and plots up to
        *max_number* (None for unlimited) halos as circles.
        """
        self.hop_output = hop_output
        self.max_number = max_number
        self.annotate = annotate
        self.min_size = min_size
        self.max_size = max_size
        self.font_size = font_size
        self.print_halo_size = print_halo_size
        self.print_halo_mass = print_halo_mass
        self.width = width

    def __call__(self, plot):
        from matplotlib.patches import Circle
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        l, b, width, height = _get_bounds(plot._axes.bbox)
        xi = lagos.x_dict[plot.data.axis]
        yi = lagos.y_dict[plot.data.axis]
        dx = plot.image._A.shape[0] / (x1-x0)
        dy = plot.image._A.shape[1] / (y1-y0)
        for halo in self.hop_output[:self.max_number]:
            size = halo.get_size()
            if size < self.min_size or size > self.max_size: continue
            # This could use halo.maximum_radius() instead of width
            if self.width is not None and \
                na.abs(halo.center_of_mass() - 
                       plot.data.center)[plot.data.axis] > \
                   self.width:
                continue
            radius = halo.maximum_radius() * dx
            center = halo.center_of_mass()
            center_x = (center[xi] - x0)*dx
            center_y = (center[yi] - y0)*dy
            cir = Circle((center_x, center_y), radius, fill=False)
            plot._axes.add_patch(cir)
            if self.annotate:
                if self.print_halo_size:
                    plot._axes.text(center_x, center_y, "%s" % size,
                    fontsize=self.font_size)
                elif self.print_halo_mass:
                    plot._axes.text(center_x, center_y, "%s" % halo.total_mass(),
                    fontsize=self.font_size)
                else:
                    plot._axes.text(center_x, center_y, "%s" % halo.id,
                    fontsize=self.font_size)

class HopParticleCallback(PlotCallback):
    _type_name = "hop_particles"
    def __init__(self, hop_output, p_size=1.0,
                max_number=None, min_size=20, alpha=0.2):
        """
        Adds particle positions for the members of each halo as identified
        by HOP. Along *axis* up to *max_number* groups in *hop_output* that are
        larger than *min_size* are plotted with *p_size* pixels per particle; 
        *alpha* determines the opacity of each particle.
        """
        self.hop_output = hop_output
        self.p_size = p_size
        self.max_number = max_number
        self.min_size = min_size
        self.alpha = alpha
    
    def __call__(self,plot):
        if self.max_number < 1: return
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        xx0, xx1 = plot._axes.get_xlim()
        yy0, yy1 = plot._axes.get_ylim()
        xf = lagos.axis_names[lagos.x_dict[plot.data.axis]]
        yf = lagos.axis_names[lagos.y_dict[plot.data.axis]]
        dx = plot.image._A.shape[0] / (x1-x0)
        dy = plot.image._A.shape[1] / (y1-y0)
        # now we loop over the haloes
        for halo in self.hop_output[:self.max_number]:
            size = halo.get_size()
            if size < self.min_size: continue
            colors = na.ones(size)
            plot._axes.hold(True)
            plot._axes.scatter(halo["particle_position_%s" % xf]*dx,
                halo["particle_position_%s" % yf]*dx, edgecolors="None",
                s=self.p_size, c='black', alpha=self.alpha)
            plot._axes.set_xlim(xx0,xx1)
            plot._axes.set_ylim(yy0,yy1)
            plot._axes.hold(False)

class VobozCircleCallback(PlotCallback):
    _type_name = "voboz_circle"
    def __init__(self, voboz_output, max_number=None,
                 annotate=False, min_size=20, font_size=8, print_halo_size=False):
        self.axis = axis
        self.voboz_output = voboz_output
        self.max_number = max_number
        self.annotate = annotate
        self.min_size = min_size
        self.font_size = font_size
        self.print_halo_size = print_halo_size

    def __call__(self, plot):
        from matplotlib.patches import Circle
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        l, b, width, height = _get_bounds(plot._axes.bbox)
        xi = lagos.x_dict[plot.data.axis]
        yi = lagos.y_dict[plot.data.axis]
        dx = plot.image._A.shape[0] / (x1-x0)
        dy = plot.image._A.shape[1] / (y1-y0)
        for i,halo in enumerate(self.voboz_output[:self.max_number]):
            if (len(halo.particles) >= self.min_size):
                radius = halo.maximum_radius * dx
                center = halo.center_of_mass
                center_x = (center[xi] - x0)*dx
                center_y = (center[yi] - y0)*dy
                #print "voboz center = (%f,%f)" % (center[xi],center[yi])
                #print "voboz radius = %f" % halo.maximum_radius
                cir = Circle((center_x, center_y), radius, fill=False)
                plot._axes.add_patch(cir)
                if self.annotate:
                    if self.print_halo_size:
                        plot._axes.text(center_x, center_y, "%s" % len(halo.particles),
                        fontsize=self.font_size)
                    else:
                        plot._axes.text(center_x, center_y, "%s" % i,
                        fontsize=self.font_size)

class CoordAxesCallback(PlotCallback):
    _type_name = "coord_axes"
    def __init__(self,unit=None,coords=False):
        """
        Creates x and y axes for a VMPlot. In the future, it will
        attempt to guess the proper units to use.
        """
        PlotCallback.__init__(self)
        self.unit = unit
        self.coords = coords

    def __call__(self,plot):
        # 1. find out what the domain is
        # 2. pick a unit for it
        # 3. run self._axes.set_xlabel & self._axes.set_ylabel to actually lay things down.
        # 4. adjust extent information to make sure labels are visable.

        # put the plot into data coordinates
        nx,ny = plot.image._A.shape
        dx = (plot.xlim[1] - plot.xlim[0])/nx
        dy = (plot.ylim[1] - plot.ylim[0])/ny

        unit_conversion = plot.pf[plot.im["Unit"]]
        aspect = (plot.xlim[1]-plot.xlim[0])/(plot.ylim[1]-plot.ylim[0])

        print "aspect ratio = ", aspect

        # if coords is False, label axes relative to the center of the
        # display. if coords is True, label axes with the absolute
        # coordinates of the region.
        xcenter = 0.
        ycenter = 0.
        if not self.coords:
            center = plot.data.center
            if plot.data.axis == 0:
                xcenter = center[1]
                ycenter = center[2]
            elif plot.data.axis == 1:
                xcenter = center[0]
                ycenter = center[2]
            else:
                xcenter = center[0]
                ycenter = center[1]


            xformat_function = lambda a,b: '%7.1e' %((a*dx + plot.xlim[0] - xcenter)*unit_conversion)
            yformat_function = lambda a,b: '%7.1e' %((a*dy + plot.ylim[0] - ycenter)*unit_conversion)
        else:
            xformat_function = lambda a,b: '%7.1e' %((a*dx + plot.xlim[0])*unit_conversion)
            yformat_function = lambda a,b: '%7.1e' %((a*dy + plot.ylim[0])*unit_conversion)
            
        xticker = matplotlib.ticker.FuncFormatter(xformat_function)
        yticker = matplotlib.ticker.FuncFormatter(yformat_function)
        plot._axes.xaxis.set_major_formatter(xticker)
        plot._axes.yaxis.set_major_formatter(yticker)
        
        xlabel = '%s (%s)' % (lagos.axis_labels[plot.data.axis][0],plot.im["Unit"])
        ylabel = '%s (%s)' % (lagos.axis_labels[plot.data.axis][1],plot.im["Unit"])
        xticksize = nx/4.
        yticksize = ny/4.
        plot._axes.xaxis.set_major_locator(matplotlib.ticker.FixedLocator([i*xticksize for i in range(0,5)]))
        plot._axes.yaxis.set_major_locator(matplotlib.ticker.FixedLocator([i*yticksize for i in range(0,5)]))
        
        plot._axes.set_xlabel(xlabel,visible=True)
        plot._axes.set_ylabel(ylabel,visible=True)
        plot._figure.subplots_adjust(left=0.1,right=0.8)

class TextLabelCallback(PlotCallback):
    _type_name = "text"
    def __init__(self, pos, text, text_args = None):
        """
        Accepts a position in (0..1, 0..1) of the image,
        some text and optionally some text arguments.
        """
        self.pos = pos
        self.text = text
        if text_args is None: text_args = {}
        self.text_args = text_args

    def __call__(self, plot):
        x = plot.image._A.shape[0] * self.pos[0]
        y = plot.image._A.shape[1] * self.pos[1]
        plot._axes.text(x, y, self.text, **self.text_args)

class ParticleCallback(PlotCallback):
    _type_name = "particles"
    region = None
    _descriptor = None
    def __init__(self, width, p_size=1.0, col='k', marker='o', stride=1.0,
                 ptype=None, stars_only=False, dm_only=False,
                 minimum_mass=None):
        """
        Adds particle positions, based on a thick slab along *axis* with a
        *width* along the line of sight.  *p_size* controls the number of
        pixels per particle, and *col* governs the color.  *ptype* will
        restrict plotted particles to only those that are of a given type.
        *minimum_mass* will require that the particles be of a given mass,
        calculated via ParticleMassMsun, to be plotted.
        """
        PlotCallback.__init__(self)
        self.width = width
        self.p_size = p_size
        self.color = col
        self.marker = marker
        self.stride = stride
        self.ptype = ptype
        self.stars_only = stars_only
        self.dm_only = dm_only
        self.minimum_mass = minimum_mass

    def __call__(self, plot):
        data = plot.data
        # we construct a recantangular prism
        x0, x1 = plot.xlim
        y0, y1 = plot.ylim
        xx0, xx1 = plot._axes.get_xlim()
        yy0, yy1 = plot._axes.get_ylim()
        reg = self._get_region((x0,x1), (y0,y1), plot.data.axis, data)
        field_x = "particle_position_%s" % lagos.axis_names[lagos.x_dict[data.axis]]
        field_y = "particle_position_%s" % lagos.axis_names[lagos.y_dict[data.axis]]
        gg = ( ( reg[field_x] >= x0 ) & ( reg[field_x] <= x1 )
           &   ( reg[field_y] >= y0 ) & ( reg[field_y] <= y1 ) )
        if self.ptype is not None:
            gg &= (reg["particle_type"] == self.ptype)
            if gg.sum() == 0: return
        if self.stars_only:
            gg &= (reg["creation_time"] > 0.0)
            if gg.sum() == 0: return
        if self.dm_only:
            gg &= (reg["creation_time"] <= 0.0)
            if gg.sum() == 0: return
        if self.minimum_mass is not None:
            gg &= (reg["ParticleMassMsun"] >= self.minimum_mass)
            if gg.sum() == 0: return
        plot._axes.hold(True)
        px, py = self.convert_to_pixels(plot,
                    [reg[field_x][gg][::self.stride],
                     reg[field_y][gg][::self.stride]])
        plot._axes.scatter(px, py, edgecolors='None', marker=self.marker,
                           s=self.p_size, c=self.color)
        plot._axes.set_xlim(xx0,xx1)
        plot._axes.set_ylim(yy0,yy1)
        plot._axes.hold(False)


    def _get_region(self, xlim, ylim, axis, data):
        LE, RE = [None]*3, [None]*3
        xax = lagos.x_dict[axis]
        yax = lagos.y_dict[axis]
        zax = axis
        LE[xax], RE[xax] = xlim
        LE[yax], RE[yax] = ylim
        LE[zax] = data.center[zax] - self.width*0.5
        RE[zax] = data.center[zax] + self.width*0.5
        if self.region is not None \
            and na.all(self.region.left_edge <= LE) \
            and na.all(self.region.right_edge >= RE):
            return self.region
        self.region = data.pf.h.periodic_region(
            data.center, LE, RE)
        return self.region

class TitleCallback(PlotCallback):
    _type_name = "title"
    def __init__(self, title="Plot"):
        """
        Accepts a *title* and adds it to the plot
        """
        PlotCallback.__init__(self)
        self.title = title

    def __call__(self,plot):
        plot._axes.set_title(self.title)




"""
Some functions to include chemistry stuff,
although it should work for generalized
tables, if fed the appropriate key.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

rates_out_key = \
      [ "tgas", \
        "k1", "k2", "k3", "k4", "k5", "k6", "k7", "k8", "k9", "k10", "k11",  \
        "k12", "k13", "k14", "k15", "k16", "k17", "k18", "k19", "k21", "k22",\
        "k23", "k50", "k51", "k52", "k53", "k54", "k55", "k56", \
        "k13_1", "k13_2", "k13_3", "k13_4", "k13_5", "k13_6", "k13_7"         ]

cool_out_key = \
      [ "tgas", \
        "ceHI", "ceHeI", "ceHeII", "ciHI", "ciHeI", "ciHeIS", "ciHeII",\
        "reHII", "reHeII1", "reHeII2", "reHeIII", "brem", "comp", \
        "gphdl", "gpldl", "cieco", "vibh", "hyd01k", "rotl", "roth", \
        "h2k01", "hdlte", "hdlow", "hdc_1", "hdc_2", "hdc_3", "hdc_4", \
        "hdc_5"]

from yt.lagos import *
#from numarray import *
#import numarray.nd_image as nd
import PointCombine
import types, exceptions

class EnzoTable:
    # This is a class for storing tables of data from enzo.
    # Specifically, we will be storing chemical rates and cooling rates
    def __init__(self, filename, key):
        # We get fed the filename and a table that is the key to the column names
        self.cols = {}
        self.invcols = {}
        for i in range(len(key)):
            self.cols[key[i]] = i
            self.invcols[i] = key[i]
        self.filename = filename
        self.readTable()
    def clearAll(self):
        try:
            del self.columns
        except:
            pass
        try:
            del self.params
        except:
            pass
    def readTable(self):
        self.clearAll()
        lines = open(self.filename).readlines()
        # We may have comment lines, which we need to strip out
        # There may be a more elegant way to do this
        # We assume that all unit lines have only one value
        # This is slow, due to the stripping and splitting, but it shouldn't be
        # a huge portion of the whole runtime
        i = 0
        self.params = {}
        toArray = []
        for line in lines:
            if line.strip()[0] == "#":
                # It's a comment/unit line
                p, v = line[1:].split("=")
                self.params[p.replace(" ","")[1:]] = float(v)
            else:
                toArray.append(map(float, line.split()))
        self.columns = na.array(toArray, 'float64')
        mylog.debug("Found %s bins of rate values", self.columns.shape[0])
    def __getitem__(self, item):
        ## This WILL get cleaned up, but it does what I want for now
        x_vals = []
        toReshape = None
        if isinstance(item, types.TupleType):
            if isinstance(item[0], types.FloatType) or \
               isinstance(item[0], types.IntType):
                x_vals.append(float(item[0]))
            elif isinstance(item[0], na.ndarray):
                toReshape = item[0].shape
                x_vals = (item[0].ravel())
            colsToReturn = []
            for col in item[1:]:
                if isinstance(col, types.StringType):
                    colsToReturn.append(self.cols[col])
                else:
                    colsToReturn.append(int(col))
        elif isinstance(item, types.FloatType) or \
             isinstance(item, types.IntType):
            x_vals.append(float(item))
            colsToReturn=arange(1,len(self.cols))
        elif isinstance(item, na.ndarray):
            toReshape = item.shape
            x_vals = item.ravel()
            colsToReturn=arange(1,len(self.cols))
        elif isinstance(item, types.ListType):
            colsToReturn=arange(1,len(self.cols))
        elif isinstance(item, types.StringType):
            return self.columns[:,self.cols[item]]
        else:
            raise exceptions.TypeError()
        colsToReturn = na.array(colsToReturn,'int32')
        valsToReturn = na.zeros((len(x_vals),len(colsToReturn)),'float64')
        x_axis = self.columns[:,0]
        x_vals_arr = na.array(x_vals, 'float64')
        PointCombine.Interpolate(x_axis, self.columns, x_vals_arr, valsToReturn, colsToReturn)
        if toReshape != None:
            if len(colsToReturn == 1):
                valsToReturn = na.reshape(valsToReturn, toReshape)
            else:
                newShape = list(toReshape)
                newShape.append(len(colsToReturn))
                valsToReturn = na.reshape(valsToReturn, newShape)
        return valsToReturn



from yt.mods import *

max_rho = []
max_pos = []
times = []
for i in range(30):
    pf = load("my_output%04i" % (i))
    v, c = pf.h.find_max("Density")
    max_rho.append(v)
    max_pos.append(c)
    times.append(pf["CosmologyCurrentTime"] * pf["years"])
pylab.loglog(times, max_rho)
pylab.clf()



from bbfreeze import Freezer
f = Freezer("yt-0.3", includes=("pytz","numpy.oldnumeric.fft","pytz.timezone"))
f.addScript("scripts/reason")
f.addScript("scripts/yt")
f()    # starts the freezing process




from yt.lagos import *
import inspect, functools

class TimeSeriesData(object):
    def __init__(self, name):
        self.outputs = []

    def __iter__(self):
        # We can make this fancier, but this works
        return self.outputs.__iter__()

class EnzoTimeSeries(TimeSeriesData):
    _enzo_header = "DATASET WRITTEN "
    def __init__(self, name, **kwargs):
        TimeSeriesData.__init__(self, name)
        output_list = kwargs.pop('output_list', None)
        output_log = kwargs.pop('output_log', None)
        if output_list: self._populate_output_list(output_list)
        if output_log: self._populate_output_log(output_log)
        for type_name in data_object_registry:
            setattr(self, type_name, functools.partial(
                TimeSeriesDataObject, self, type_name))

    def _populate_output_list(self, output_list):
        for output in output_list:
            self._insert(EnzoStaticOutput(output))

    def _populate_output_log(self, output_log):
        for line in open(output_log):
            if not line.startswith(_enzo_header): continue
            fn = line[len(_enzo_header):].strip()
            self._insert(EnzoStaticOutput(fn))

    def __getitem__(self, key):
        if isinstance(key, types.SliceType):
            if isinstance(key.start, types.FloatType):
                return self.get_range(key.start, key.stop)
        return self.outputs[key]
        
    def _insert(self, pf):
        # We get handed an instantiated parameter file
        # Here we'll figure out a couple things about it, and then stick it
        # inside our list.
        self.outputs.append(pf)
        
    def eval(self, tasks, obj=None):
        if obj == None: obj = TimeSeriesDataObject(self, "all_data")
        tasks = ensure_list(tasks)
        return_values = []
        for pf in self:
            return_values.append([])
            for task in tasks:
                style = inspect.getargspec(task.eval)[0][1]
                if style == 'pf': arg = pf
                elif style == 'data_object': arg = obj.get(pf)
                return_values[-1].append(task.eval(arg))
        return return_values

class TimeSeriesDataObject(object):
    def __init__(self, time_series, data_object_name, *args, **kwargs):
        self.time_series = weakref.proxy(time_series)
        self.data_object_name = data_object_name
        self._args = args
        self._kwargs = kwargs

    def eval(self, tasks):
        return self.time_series.eval(tasks, self)

    def get(self, pf):
        # We get the type name, which corresponds to an attribute of the
        # hierarchy
        cls = getattr(pf.h, self.data_object_name)
        return cls(*self._args, **self._kwargs)



"""
Import the components of the volume rendering extension

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as na

from transfer_functions import TransferFunction, ColorTransferFunction, \
                             PlanckTransferFunction, \
                             MultiVariateTransferFunction, \
                             ProjectionTransferFunction
from yt.amr_utils import PartitionedGrid, VectorPlane, \
                             TransferFunctionProxy
from grid_partitioner import HomogenizedVolume, \
                             HomogenizedBrickCollection, \
                             export_partitioned_grids, \
                             import_partitioned_grids
from image_handling import export_rgba, import_rgba, \
                           plot_channel, plot_rgb
from software_sampler import VolumeRendering
from camera import Camera, PerspectiveCamera, StereoPairCamera



"""
Test that we can get outputs, and interact with them in some primitive ways.
"""

# @TODO: Add unit test for deleting field from FieldInfo

import unittest, glob, os.path, os, sys, StringIO

print "Reporting from %s" % (os.getcwd())
sys.path = ['.'] + sys.path

from yt.config import ytcfg
ytcfg["yt","LogLevel"] = '50'
ytcfg["yt","logFile"] = "False"
ytcfg["yt","suppressStreamLogging"] = "True"
ytcfg["lagos","serialize"] = "False"

import cPickle
import yt.lagos
import yt.lagos.OutputTypes
import numpy as na
from yt.fido import ParameterFileStore

# The dataset used is located at:
# http://yt.spacepope.org/DD0018.zip
fn = "DD0010/moving7_0010"
fn = os.path.join(os.path.dirname(__file__), fn)

class LagosTestingBase:
    def setUp(self):
        self.OutputFile = yt.lagos.EnzoStaticOutput(fn)
        self.hierarchy = self.OutputFile.hierarchy
        self.v, self.c = self.hierarchy.find_max("Density")
        gp = os.path.join(os.path.dirname(fn),"*.yt")
        ytFiles = glob.glob(gp)
        for i in ytFiles:
            #print "Removing %s" % (i)
            os.unlink(i)

    def tearDown(self):
        if hasattr(self,'data'): del self.data
        if hasattr(self,'region'): del self.region
        if hasattr(self,'ind_to_get'): del self.ind_to_get
        del self.OutputFile, self.hierarchy
        
class TestParameterFileStore(unittest.TestCase):
    def setUp(self):
        self.original = (yt.config.ytcfg.get("yt","ParameterFileStore"),
                         yt.config.ytcfg.get("lagos","serialize"))
        ytcfg['yt','ParameterFileStore'] = "testing.csv"
        pfs = ParameterFileStore()
        os.unlink(pfs._get_db_name())
        self.pfs = ParameterFileStore() # __init__ gets called again
        ytcfg['lagos', 'serialize'] = "True"

    def testCacheFile(self):
        pf1 = yt.lagos.EnzoStaticOutput(fn)
        pf2 = self.pfs.get_pf_hash(pf1._hash())
        self.assertTrue(pf1 is pf2)

    def testGrabFile(self):
        pf1 = yt.lagos.EnzoStaticOutput(fn)
        hash = pf1._hash()
        del pf1
        pf2 = self.pfs.get_pf_hash(hash)
        self.assertTrue(hash == pf2._hash())

    def testGetCurrentTimeID(self):
        pf1 = yt.lagos.EnzoStaticOutput(fn)
        hash = pf1._hash()
        ctid = pf1["CurrentTimeIdentifier"]
        del pf1
        pf2 = self.pfs.get_pf_ctid(ctid)
        self.assertTrue(hash == pf2._hash())

    def tearDown(self):
        os.unlink(self.pfs._get_db_name())
        ytcfg['yt', 'ParameterFileStore'] = self.original[0]
        ytcfg['lagos', 'serialize'] = self.original[1]
        self.pfs.__init__()

class TestHierarchy(LagosTestingBase, unittest.TestCase):
    def testGetHierarchy(self):
        self.assert_(self.OutputFile.hierarchy != None)

    def testGetUnits(self):
        self.assert_(self.OutputFile["cm"] != 1.0)

    def testGetSmallestDx(self):
        self.assertAlmostEqual(self.hierarchy.get_smallest_dx(),
                               0.00048828125, 7)

    def testGetNumberOfGrids(self):
        self.assertEqual(self.hierarchy.num_grids, len(self.hierarchy.grids))
        self.assertEqual(self.hierarchy.num_grids, 10)

    def testChildrenOfRootGrid(self):
        for child in self.hierarchy.grids[0].Children:
            self.assert_(child.Parent.id == self.hierarchy.grids[0].id)

    def testGetSelectLevels(self):
        for level in range(self.hierarchy.max_level+1):
            for grid in self.hierarchy.select_grids(level):
                self.assert_(grid.Level == level)

    def testPrintStats(self):
        a = sys.stdout
        sys.stdout = StringIO.StringIO()
        try:
            self.hierarchy.print_stats()
            worked = True
        except:
            worked = False
        sys.stdout = a
        self.assert_(worked)

    def testDataTypes(self):
        r=self.hierarchy.region(
                     [0.5,0.5,0.5],[0.0, 0.0, 0.0],
                     [1.0, 1.0, 1.0],
                     ["CellMass","Temperature"])
            # Testing multiple fields fed in
        s=self.hierarchy.sphere(
                     [0.5,0.5,0.5],2.0,
                     ["CellMass","Temperature"])
        ms = s["CellMass"].sum() # Testing adding new field transparently
        mr = r["CellMass"].sum() # Testing adding new field transparently
        self.assertEqual(ms,mr)  # Asserting equality between the two

    def testProjectionCorrectnessMultipleFields(self):
        p = self.hierarchy.proj(0,["Density","Ones"], weight=None) # Unweighted
        self.assertTrue(na.all(p["Ones"] == 1.0))

    def testProjectionMakingMultipleFields(self):
        p = self.hierarchy.proj(0,["Density","Temperature","Ones"],weight_field="Ones") # Unweighted
        # One for each field, pdx, pdy, px, py, and one for the weight
        self.assertEqual(len(p.data.keys()), 8)

    def testProjectionSuccess(self):
        p = self.hierarchy.proj(0,"Density") # Unweighted
        p = self.hierarchy.proj(1,"Temperature","Density") # Weighted
        p = self.hierarchy.proj(2,"Entropy") # Derived field

    def testUnweightedProjectionCorrectness(self):
        # Now we test that we get good answers
        for axis in range(3):
            p = self.hierarchy.proj(axis, "Ones") # Derived field
            self.assertTrue(na.all(p["Ones"] == 1.0))
            # Regardless of weighting, we want ones back

    def testWeightedProjectionCorrectness(self):
        # Now we test that we get good answers
        for axis in range(3):
            # Regardless of weighting, we want ones back
            p = self.hierarchy.proj(axis, "Ones", "Density")
            self.assertTrue(na.all(p["Ones"] == 1.0))

# Now we test each datatype in turn

def _returnFieldFunction(field):
    def field_function(self):
        try:
            self.data[field.name]
            if not field.particle_type and not field.vector_field and \
                self.data[field.name].size > 1:
                self.assertEqual(na.product(self.data["Density"].shape),
                                 na.product(self.data[field.name].shape))
            del self.data[field.name]
        except yt.lagos.ValidationException:
            pass
    return field_function

def _returnProfile1DFunction(field, weight, accumulation, lazy):
    def add_field_function(self):
        self.data.set_field_parameter("center",[.5,.5,.5])
        profile = yt.lagos.BinnedProfile1D(
            self.data, 8, "RadiusCode", 0, 1.0, False, lazy)
        profile.add_fields(field, weight=weight, accumulation=accumulation)
    return add_field_function

def _returnProfile2DFunction(field, weight, accumulation, lazy):
    def add_field_function(self):
        self.data.set_field_parameter("center",[.5,.5,.5])
        cv_min = self.hierarchy.get_smallest_dx()**3.0
        cv_max = 1.0 / max(self.OutputFile["TopGridDimensions"])
        profile = yt.lagos.BinnedProfile2D(self.data,
                    8, "RadiusCode", 1e-3, 1.0, True,
                    8, "CellVolumeCode", cv_min, cv_max, True, lazy)
        profile.add_fields(field, weight=weight, accumulation=accumulation)
    return add_field_function

class DataTypeTestingBase:
    def setUp(self):
        LagosTestingBase.setUp(self)

    def testRepr(self):
        self.assertTrue(
            ("%s" % self.data).startswith(self.data.__class__.__name__))

class Data3DBase:
    def testProfileAccumulateMass(self):
        self.data.set_field_parameter("center",[0.5]*3)
        profile = yt.lagos.BinnedProfile1D(self.data, 8, "RadiusCode", 0, 1.0,
                                           False, True)
        profile.add_fields("CellMassMsun", weight=None, accumulation=True)
        v1 = profile["CellMassMsun"].max()
        v2 = self.data["CellMassMsun"].sum()
        v2 = na.abs(1.0 - v2/v1)
        self.assertAlmostEqual(v2, 0.0, 7)

    def testExtractConnectedSetsNoCache(self):
        mi = self.data["Density"].min() * 2.0
        ma = self.data["Density"].max() * 0.99
        cons, contours = self.data.extract_connected_sets(
            "Density", 2, mi, ma)
        self.assertEqual(len(contours), 2) # number of contour levels
        self.assertEqual(len(contours[0]), 2)
        self.assertEqual(len(contours[1]), 1)

    def testExtractConnectedSetsCache(self):
        mi = self.data["Density"].min() * 2.0
        ma = self.data["Density"].max() * 0.99
        cons, contours = self.data.extract_connected_sets(
            "Density", 2, mi, ma, cache=True)
        self.assertEqual(len(contours), 2) # number of contour levels
        self.assertEqual(len(contours[0]), 2)
        self.assertEqual(len(contours[1]), 1)

    def testContoursCache(self):
        cid = yt.lagos.identify_contours(self.data, "Density",
                self.data["Density"].min()*2.00,
                self.data["Density"].max()*1.01)
        self.assertEqual(len(cid), 2)

    def testContoursObtain(self):
        cid = yt.lagos.identify_contours(self.data, "Density",
                self.data["Density"].min()*2.00, self.data["Density"].max()*1.01)
        self.assertEqual(len(cid), 2)

    def testContoursValidityMax(self):
        v1 = self.data["Density"].max()*0.99
        v2 = self.data["Density"].max()*1.01
        cid = yt.lagos.identify_contours(self.data, "Density", v1, v2)
        self.assertTrue(na.all(v1 < self.data["Density"][cid[0]])
                    and na.all(v2 > self.data["Density"][cid[0]]))
        self.assertEqual(len(cid), 1)

    def testContoursValidityMin(self):
        v1 = self.data["Density"].min()*0.99
        v2 = self.data["Density"].min()*1.01
        cid = yt.lagos.identify_contours(self.data, "Density", v1, v2)
        self.assertTrue(na.all(v1 < self.data["Density"][cid[0]])
                    and na.all(v2 > self.data["Density"][cid[0]]))
        self.assertEqual(len(cid), 3)

    def testPickle(self):
        ps = cPickle.dumps(self.data)
        pf, obj = cPickle.loads(ps)
        self.assertEqual(obj["CellMassMsun"].sum(), self.data["CellMassMsun"].sum())

for field_name in yt.lagos.FieldInfo:
    if field_name.startswith("PT"): continue
    field = yt.lagos.FieldInfo[field_name]
    setattr(DataTypeTestingBase, "test%s" % field.name, _returnFieldFunction(field))

field = "Temperature"
for weight in [None, "CellMassMsun"]:
    for lazy in [True, False]:
        for accumulation in [True, False]:
            func = _returnProfile1DFunction(field, weight, accumulation, lazy)
            name = "test%sProfile1D_w%s_l%s_a%s" % (field,
                                                weight, lazy,
                                                accumulation)
            setattr(Data3DBase, name, func)

for weight in [None, "CellMassMsun"]:
    for lazy in [True, False]:
        for accumulation_x in [True, False]:
            for accumulation_y in [True, False]:
                acc = (accumulation_x, accumulation_y)
                func = _returnProfile2DFunction(field, weight, acc, lazy)
                name = "test%sProfile2D_w%s_l%s_a%s_a%s" % (field,
                                                        weight, lazy,
                                                        accumulation_x,
                                                        accumulation_y)
                setattr(Data3DBase, name, func)

class TestSmoothedCoveringGrid(LagosTestingBase, unittest.TestCase):
    def setUp(self):
        LagosTestingBase.setUp(self)

    def testAllCover(self):
        DIMS = 32
        for i in range(self.hierarchy.max_level+1):
            dx = (DIMS*2**i)**-1
            LE = na.array([0.5,0.5,0.5])-(dx*DIMS/2.0)
            RE = na.array([0.5,0.5,0.5])+(dx*DIMS/2.0)
            cg = self.hierarchy.smoothed_covering_grid(
                    level=i, left_edge=LE, right_edge=RE,
                    dims=[DIMS]*3, fields=["Density"])
            self.assertFalse(na.any(na.isnan(cg["Density"])))
            self.assertFalse(na.any(cg["Density"]==-999))

    def testAddField(self):
        DIMS = 64
        i = 5
        dx = (DIMS*2**i)**-1
        LE = na.array([0.5,0.5,0.5])-(dx*DIMS/2.0)
        RE = na.array([0.5,0.5,0.5])+(dx*DIMS/2.0)
        cg = self.hierarchy.smoothed_covering_grid(
                level=i, left_edge=LE, right_edge=RE,
                dims=[DIMS]*3, fields=["Density"])
        self.assertFalse(na.any(na.isnan(cg["Temperature"])))
        self.assertFalse(na.any(cg["Temperature"]==-999))

class TestIntSmoothedCoveringGrid(LagosTestingBase, unittest.TestCase):
    def setUp(self):        
        LagosTestingBase.setUp(self)

    def testCoordinates(self):
        # We skip the first grid because it has ghost zones on all sides
        for g in self.hierarchy.grids[1:]:
            LE = g.LeftEdge - g.dds
            level = g.Level
            dims = g.ActiveDimensions + 2
            g1 = self.hierarchy.si_covering_grid(level, LE, dims)
            g2 = g.retrieve_ghost_zones(1, ["x","y","z"], smoothed=False)
            for field in 'xyz':
                diff = na.abs((g1[field] - g2[field])/(g1[field] + g2[field]))
                self.assertAlmostEqual(diff.max(), 0.0, 1e-14)

class TestDataCube(LagosTestingBase, unittest.TestCase):
    def setUp(self):
        LagosTestingBase.setUp(self)

    def testNoGhost(self):
        DW = self.OutputFile["DomainRightEdge"] \
           - self.OutputFile["DomainLeftEdge"]
        for g in self.hierarchy.grids:
            cube = g.retrieve_ghost_zones(0, "Density")
            self.assertTrue(na.all(cube["Density"] == g["Density"]))
            cube["Density"] = na.arange(cube["Density"].size).reshape(cube["Density"].shape)
            cube.flush_data(field="Density")
            self.assertTrue(na.all(g["Density"] == cube["Density"]))

    def testOffsetDomain(self):
        DW = self.OutputFile["DomainRightEdge"] \
           - self.OutputFile["DomainLeftEdge"]
        for g in self.hierarchy.grids:
            cube = self.hierarchy.covering_grid(g.Level,
                g.LeftEdge+DW, g.ActiveDimensions)
            self.assertTrue(na.all(g["Density"] == cube["Density"]))

    def testTwoGhost(self):
        for g in self.hierarchy.grids:
            cube = g.retrieve_ghost_zones(2, "Density")

    def testMultipleFields(self):
        for g in self.hierarchy.grids:
            cube1 = g.retrieve_ghost_zones(0, ["Density","Temperature"])
            self.assertTrue(na.all(cube1["Density"] == g["Density"]))
            self.assertTrue(na.all(cube1["Temperature"] == g["Temperature"]))
            cube2a = g.retrieve_ghost_zones(0, "Density")
            cube2b = g.retrieve_ghost_zones(0, "Temperature")
            self.assertTrue(na.all(cube1["Density"] == cube2a["Density"]))
            self.assertTrue(na.all(cube1["Temperature"] == cube2b["Temperature"]))
    
    def testFlushBackToGrids(self):
        ml = self.hierarchy.max_level
        cg = self.hierarchy.covering_grid(2, [0.0]*3, [64,64,64])
        cg["Ones"] *= 2.0
        cg.flush_data(field="Ones")
        for g in na.concatenate([self.hierarchy.select_grids(i) for i in range(3)]):
            self.assertEqual(g["Ones"].max(), 2.0)
            self.assertEqual(g["Ones"][g["Ones"]*g.child_mask>0].min(), 2.0)

    def testFlushBackToNewCover(self):
        ml = self.hierarchy.max_level
        cg = self.hierarchy.covering_grid(2, [0.0]*3, [64,64,64])
        cg["tempContours"] = cg["Ones"] * 2.0
        cg.flush_data(field="tempContours")
        cg2 = self.hierarchy.covering_grid(2, [0.0]*3, [64,64,64])
        self.assertTrue(na.all(cg["tempContours"] == cg2["tempContours"]))

    def testRawFlushBack(self):
        ml = self.hierarchy.max_level
        cg = self.hierarchy.covering_grid(2, [0.0]*3, [64,64,64])
        cg["DensityNew"] = cg["Density"] * 2.111
        cg.flush_data(field="DensityNew")
        for g in na.concatenate([self.hierarchy.select_grids(i) for i in range(3)]):
            ni = g["DensityNew"] > 0
            min_diff = (g["DensityNew"][ni]/g["Density"][ni]).max()
            max_diff = (g["DensityNew"][ni]/g["Density"][ni]).min()
            min_diff_i = na.argmin(g["DensityNew"][ni]/g["Density"][ni])
            max_diff_i = na.argmax(g["DensityNew"][ni]/g["Density"][ni])
            self.assertAlmostEqual(min_diff, 2.111, 5)
            self.assertAlmostEqual(max_diff, 2.111, 5)

    def testAllCover(self):
        cg = self.hierarchy.covering_grid(1, [0.0]*3, [32,32,32])
        mi, ma = 1e30, -1e30
        for g in na.concatenate([self.hierarchy.select_grids(i) for i in range(2)]):
            ma = max(ma, g["Density"].max())
            mi = min(mi, g["Density"].min())
        self.assertEqual(cg["Density"].max(), ma)
        self.assertEqual(cg["Density"].min(), mi)

    def testCellVolume(self):
        cg = self.hierarchy.covering_grid(2, [0.0]*3, [64,64,64])
        self.assertEqual(na.unique(cg["CellVolume"]).size, 1)


class TestDiskDataType(Data3DBase, DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data=self.hierarchy.disk(
                     [0.5,0.5,0.5],[0.2, 0.1, 0.5],1.0,1.0)

class TestRegionDataType(Data3DBase, DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data=self.hierarchy.region(
                     [0.5,0.5,0.5],[0.0, 0.0, 0.0],
                     [1.0, 1.0, 1.0])
    def testVolume(self):
        vol = self.data["CellVolume"].sum() / self.data.convert("cm")**3.0
        self.assertAlmostEqual(vol,1.0,7)

class TestRegionStrictDataType(Data3DBase, DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data=self.hierarchy.region_strict(
                     [0.5,0.5,0.5],[0.0, 0.0, 0.0],
                     [1.0, 1.0, 1.0])
    def testVolume(self):
        vol = self.data["CellVolume"].sum() / self.data.convert("cm")**3.0
        self.assertAlmostEqual(vol,1.0,7)

class TestPeriodicRegionDataType(Data3DBase, DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data=self.hierarchy.periodic_region(
                     [0.5,0.5,0.5],[0.5, 0.5, 0.5],
                     [1.5,1.5,1.5])
    def testVolume(self):
        vol = self.data["CellVolume"].sum() / self.data.convert("cm")**3.0
        self.assertAlmostEqual(vol,1.0,7)

class TestPeriodicRegionStrictDataType(Data3DBase,
            DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data=self.hierarchy.periodic_region_strict(
                     [0.5,0.5,0.5],[0.5, 0.5, 0.5],
                     [1.5,1.5,1.5])
    def testVolume(self):
        vol = self.data["CellVolume"].sum() / self.data.convert("cm")**3.0
        self.assertAlmostEqual(vol,1.0,7)

class TestSphereDataType(Data3DBase, DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data=self.hierarchy.sphere([0.5,0.5,0.5],1.0)
    def testVolume(self):
        vol = self.data["CellVolume"].sum() / self.data.convert("cm")**3.0
        self.assertAlmostEqual(vol,1.0,7)

class TestSliceDataType(DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data = self.hierarchy.slice(0,0.5, center=[0.5, 0.5, 0.5])

class TestCuttingPlane(DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data = self.hierarchy.cutting([0.1,0.3,0.4], [0.5,0.5,0.5], ["Density"])
    def testAxisVectors(self):
        x_v = self.data._x_vec
        y_v = self.data._y_vec
        z_v = self.data._norm_vec
        self.assertAlmostEqual(na.dot(x_v, y_v), 0.0, 7)
        self.assertAlmostEqual(na.dot(x_v, z_v), 0.0, 7)
        self.assertAlmostEqual(na.dot(y_v, z_v), 0.0, 7)
    def testZHeight(self):
        self.assertTrue(na.all(self.data['pz'] < self.data['dx']))

class TestGridDataType(DataTypeTestingBase, LagosTestingBase, unittest.TestCase):
    def setUp(self):
        DataTypeTestingBase.setUp(self)
        self.data = self.hierarchy.grids[0]

class TestExtractFromSphere(TestSphereDataType):
    def setUp(self):
        TestSphereDataType.setUp(self)
        self.region = self.data
        self.ind_to_get = na.where(self.region["Temperature"]>500)
        self.data = self.region.extract_region(self.ind_to_get)

    def testNumberOfEntries(self):
        self.assertEqual(self.ind_to_get[0].shape,
                        self.data["Density"].shape)
    def testVolume(self):
        self.ind_to_get = na.where(self.region["CellVolume"]>0.0)
        vol = self.region.extract_region(self.ind_to_get)["CellVolume"].sum() \
            / self.data.convert("cm")**3.0
        self.assertAlmostEqual(vol,1.0,7)

    def testJoin(self):
        new_region = self.region.extract_region(
                self.region["Temperature"]<=500)
        joined_region = self.data.join(new_region)
        self.assertEqual(joined_region["CellMassMsun"].sum(),
                         self.region["CellMassMsun"].sum())

    # I have verified that the *old* version of the code overconnected
    # contours outside.  This are overridden to make sure this does not
    # happen again!
    def testContoursObtain(self):
        cid = yt.lagos.identify_contours(self.data, "Density",
                self.data["Density"].min()*2.00, self.data["Density"].max()*1.01)
        self.assertEqual(len(cid), 10)

    def testContoursCache(self):
        cid = yt.lagos.identify_contours(self.data, "Density",
                self.data["Density"].min()*2.00,
                self.data["Density"].max()*1.01)
        self.assertEqual(len(cid), 10)

    def testExtractConnectedSetsNoCache(self):
        mi = self.data["Density"].min() * 2.0
        ma = self.data["Density"].max() * 0.99
        cons, contours = self.data.extract_connected_sets(
            "Density", 2, mi, ma)
        self.assertEqual(len(contours), 2) # number of contour levels
        self.assertEqual(len(contours[0]), 10)
        self.assertEqual(len(contours[1]), 1)

    def testExtractConnectedSetsCache(self):
        mi = self.data["Density"].min() * 2.0
        ma = self.data["Density"].max() * 0.99
        cons, contours = self.data.extract_connected_sets(
            "Density", 2, mi, ma, cache=True)
        self.assertEqual(len(contours), 2) # number of contour levels
        self.assertEqual(len(contours[0]), 10)
        self.assertEqual(len(contours[1]), 1)

class TestExtractFromRegion(TestRegionDataType):
    def setUp(self):
        TestRegionDataType.setUp(self)
        self.region = self.data
        self.ind_to_get = na.where(self.region["Temperature"]>500)
        self.data = self.region.extract_region(self.ind_to_get)
    def testNumberOfEntries(self):
        self.assertEqual(self.ind_to_get[0].shape,
                        self.data["Density"].shape)
    def testVolume(self):
        ind_to_get = na.where(self.region["CellVolume"]>0.0)
        vol = self.region.extract_region(ind_to_get)["CellVolume"].sum() \
            / self.data.convert("cm")**3.0
        self.assertAlmostEqual(vol,1.0,7)

    def testJoin(self):
        new_region = self.region.extract_region(
                self.region["Temperature"]<=500)
        joined_region = self.data.join(new_region)
        self.assertEqual(joined_region["CellMassMsun"].sum(),
                         self.region["CellMassMsun"].sum())

    # I have verified that the *old* version of the code overconnected
    # contours outside.  This are overridden to make sure this does not
    # happen again!
    def testContoursObtain(self):
        cid = yt.lagos.identify_contours(self.data, "Density",
                self.data["Density"].min()*2.00, self.data["Density"].max()*1.01)
        self.assertEqual(len(cid), 10)

    def testContoursCache(self):
        cid = yt.lagos.identify_contours(self.data, "Density",
                self.data["Density"].min()*2.00,
                self.data["Density"].max()*1.01)
        self.assertEqual(len(cid), 10)

    def testExtractConnectedSetsNoCache(self):
        mi = self.data["Density"].min() * 2.0
        ma = self.data["Density"].max() * 0.99
        cons, contours = self.data.extract_connected_sets(
            "Density", 2, mi, ma)
        self.assertEqual(len(contours), 2) # number of contour levels
        self.assertEqual(len(contours[0]), 10)
        self.assertEqual(len(contours[1]), 1)

    def testExtractConnectedSetsCache(self):
        mi = self.data["Density"].min() * 2.0
        ma = self.data["Density"].max() * 0.99
        cons, contours = self.data.extract_connected_sets(
            "Density", 2, mi, ma, cache=True)
        self.assertEqual(len(contours), 2) # number of contour levels
        self.assertEqual(len(contours[0]), 10)
        self.assertEqual(len(contours[1]), 1)


class TestUnilinearInterpolator(unittest.TestCase):
    def setUp(self):
        x0, x1 = na.random.uniform(-100,100,2)
        nstep_x = na.random.randint(10,200)
        nvals = na.random.randint(100,1000)

        table = na.mgrid[x0:x1:nstep_x*1j]

        self.ufi_x = yt.lagos.UnilinearFieldInterpolator(table,
                      (x0,x1),'x')
        self.my_dict = {}
        self.my_dict['x'] = na.random.uniform(x0,x1,nvals)

    def testXInt(self):
        nv = self.ufi_x(self.my_dict)
        for i,v in enumerate(nv):
            self.assertAlmostEqual(v, self.my_dict['x'][i], 5)

class TestBilinearInterpolator(unittest.TestCase):
    def setUp(self):
        x0, x1 = na.random.uniform(-100,100,2)
        y0, y1 = na.random.uniform(-100,100,2)
        nstep_x = na.random.randint(10,200)
        nstep_y = na.random.randint(10,200)
        nvals = na.random.randint(100,1000)

        table = na.mgrid[x0:x1:nstep_x*1j,
                         y0:y1:nstep_y*1j]

        self.bfi_x = yt.lagos.BilinearFieldInterpolator(table[0,...],
                      (x0,x1,y0,y1),['x','y'])
        self.bfi_y = yt.lagos.BilinearFieldInterpolator(table[1,...],
                      (x0,x1,y0,y1),['x','y'])
        self.my_dict = {}
        self.my_dict['x'] = na.random.uniform(x0,x1,nvals)
        self.my_dict['y'] = na.random.uniform(y0,y1,nvals)

    def testXInt(self):
        nv = self.bfi_x(self.my_dict)
        for i,v in enumerate(nv):
            self.assertAlmostEqual(v, self.my_dict['x'][i], 5)

    def testYInt(self):
        nv = self.bfi_y(self.my_dict)
        for i,v in enumerate(nv):
            self.assertAlmostEqual(v, self.my_dict['y'][i], 5)

class TestTrilinearInterpolator(unittest.TestCase):
    def setUp(self):
        x0, x1 = na.random.uniform(-100,100,2)
        y0, y1 = na.random.uniform(-100,100,2)
        z0, z1 = na.random.uniform(-100,100,2)
        nstep_x = na.random.randint(10,200)
        nstep_y = na.random.randint(10,200)
        nstep_z = na.random.randint(10,200)
        nvals = na.random.randint(100,1000)

        table = na.mgrid[x0:x1:nstep_x*1j,
                         y0:y1:nstep_y*1j,
                         z0:z1:nstep_z*1j]

        self.tfi_x = yt.lagos.TrilinearFieldInterpolator(table[0,...],
                      (x0,x1,y0,y1,z0,z1),['x','y','z'])
        self.tfi_y = yt.lagos.TrilinearFieldInterpolator(table[1,...],
                      (x0,x1,y0,y1,z0,z1),['x','y','z'])
        self.tfi_z = yt.lagos.TrilinearFieldInterpolator(table[2,...],
                      (x0,x1,y0,y1,z0,z1),['x','y','z'])
        self.my_dict = {}
        self.my_dict['x'] = na.random.uniform(x0,x1,nvals)
        self.my_dict['y'] = na.random.uniform(y0,y1,nvals)
        self.my_dict['z'] = na.random.uniform(z0,z1,nvals)

    def testXInt(self):
        nv = self.tfi_x(self.my_dict)
        for i,v in enumerate(nv):
            self.assertAlmostEqual(v, self.my_dict['x'][i], 5)

    def testYInt(self):
        nv = self.tfi_y(self.my_dict)
        for i,v in enumerate(nv):
            self.assertAlmostEqual(v, self.my_dict['y'][i], 5)

    def testZInt(self):
        nv = self.tfi_z(self.my_dict)
        for i,v in enumerate(nv):
            self.assertAlmostEqual(v, self.my_dict['z'][i], 5)



if __name__ == "__main__":
    unittest.main()



"""
The slowest ray-tracer known to man.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.lagos import *

class SlowRayTracer(object):
    def __init__(self, pf, axis, bounds, res, func, reverse = False):
        self.pf = pf
        self.axis = axis
        self.x_bounds, self.y_bounds = bounds
        self.res = res
        self.func = func
        self.reverse = reverse
        self.buff = na.zeros(res, dtype='float64')
        self.fill_buff()

    def _trace_single_ray(self, x, y):
        ray = self.pf.h.ortho_ray(self.axis, (x, y))
        order = ray[axis_names[self.axis]].argsort()
        if self.reverse: order = order[::-1]
        return self.func(ray, order)

    def fill_buff(self):
        #pb = get_pbar("Ray tracing ", na.prod(self.res))
        xs = na.linspace(self.x_bounds[0], self.x_bounds[1], self.res[0])
        ys = na.linspace(self.y_bounds[0], self.y_bounds[1], self.res[1])
        #counter = 0
        for i, x in enumerate(xs.ravel()):
            #counter += 1
            print i
            for j, y in enumerate(ys.ravel()):
                #pb.update(counter)
                self.buff[i,j] = self._trace_single_ray(x,y)
        #pb.finish()

    def __getitem__(self, item):
        return self.buff[item]

if __name__ == "__main__":
    def raw_proj(ray, order):
        return na.sum((ray["Density"]*ray["dx"])[order])
        #return na.sum((ray["Density"]*ray["dx"])[order])
    import yt.logger
    yt.logger.disable_stream_logging()
    pf = EnzoStaticOutput("/Users/matthewturk/Research/data/galaxy1200.dir/galaxy1200")
    srt = SlowRayTracer(pf, 0, ((0.0125,1.0-0.0125),(0.0125,1.0-0.0125)),
                        (256, 256), raw_proj, reverse=False)
    import pylab
    pylab.imshow(na.log10(srt.buff), interpolation='nearest', origin='lower')
    pylab.savefig("srt_raw_x.png")
    pylab.clf()



"""
This is a library of yt-defined exceptions

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# We don't need to import 'exceptions'
#import exceptions

class YTException(Exception):
    def __init__(self, pf = None):
        Exception.__init__(self)
        self.pf = pf

# Data access exceptions:

class YTSphereTooSmall(YTException):
    def __init__(self, pf, radius, smallest_cell):
        YTException.__init__(self, pf)
        self.radius = radius
        self.smallest_cell = smallest_cell

    def __str__(self):
        return "%0.5e < %0.5e" % (self.radius, self.smallest_cell)



from yt.logger import lagosLogger as mylog
import yt.lagos as lagos
import yt.funcs as funcs
import numpy as na
import os

dt_Tolerance = 1e-3

class EnzoSimulation(object):
    """
    Super class for performing the same operation over all data dumps in 
    a simulation from one redshift to another.
    """
    def __init__(self, EnzoParameterFile, initial_time=None, final_time=None, initial_redshift=None, final_redshift=None,
                 links=False, enzo_parameters=None, get_time_outputs=True, get_redshift_outputs=True, get_available_data=False):
        """
        Initialize an EnzoSimulation object.
        :param initial_time (float): the initial time in code units for the dataset list.  Default: None.
        :param final_time (float): the final time in code units for the dataset list.  Default: None.
        :param initial_redshift (float): the initial (highest) redshift for the dataset list.  Only for 
               cosmological simulations.  Default: None.
        :param final_redshift (float): the final (lowest) redshift for the dataset list.  Only for cosmological 
               simulations.  Default: None.
        :param links (bool): if True, each entry in the dataset list will contain entries, previous and next, that 
               point to the previous and next entries on the dataset list.  Default: False.
        :param enzo_parameters (dict): a dictionary specify additional parameters to be retrieved from the 
               parameter file.  The format should be the name of the parameter as the key and the variable type as 
               the value.  For example, {'CosmologyComovingBoxSize':float}.  All parameter values will be stored in 
               the dictionary attribute, enzoParameters.  Default: None.
        :param get_time_outputs (bool): if False, the time datasets, specified in Enzo with the dtDataDump, will not 
               be added to the dataset list.  Default: True.
        :param get_redshift_outputs (bool): if False, the redshift datasets will not be added to the dataset list.  Default: True.
        :param get_available_data (bool): if True, only datasets that are found to exist at the file path are added 
               to the list.  Devault: False.
        """
        self.EnzoParameterFile = EnzoParameterFile
        self.enzoParameters = {}
        self.redshiftOutputs = []
        self.timeOutputs = []
        self.allOutputs = []
        self.InitialTime = initial_time
        self.FinalTime = final_time
        self.InitialRedshift = initial_redshift
        self.FinalRedshift = final_redshift
        self.links = links
        self.get_time_outputs = get_time_outputs
        self.get_redshift_outputs = get_redshift_outputs
        self.get_available_data = get_available_data

        # Add any extra parameters to parameter dict.
        if enzo_parameters is None: enzo_parameters = {}
        EnzoParameterDict.update(enzo_parameters)

        # Set some parameter defaults.
        self._SetParameterDefaults()

        # Read parameters.
        self._ReadEnzoParameterFile()

        # Check for sufficient starting/ending parameters.
        if self.InitialTime is None and self.InitialRedshift is None:
            if self.enzoParameters['ComovingCoordinates'] and \
               'CosmologyInitialRedshift' in self.enzoParameters:
                self.InitialRedshift = self.enzoParameters['CosmologyInitialRedshift']
            elif 'InitialTime' in self.enzoParameters:
                self.InitialTime = self.enzoParameters['InitialTime']
            else:
                mylog.error("Couldn't find parameter for initial time or redshift from parameter file.")
                return None

        if self.FinalTime is None and self.FinalRedshift is None:
            if self.enzoParameters['ComovingCoordinates'] and \
               'CosmologyFinalRedshift' in self.enzoParameters:
                self.FinalRedshift = self.enzoParameters['CosmologyFinalRedshift']
            elif 'StopTime' in self.enzoParameters:
                self.FinalTime = self.enzoParameters['StopTime']
            else:
                mylog.error("Couldn't find parameter for final time or redshift from parameter file.")
                return None

        # Convert initial/final redshifts to times.
        if self.enzoParameters['ComovingCoordinates']:
            # Instantiate a cosmology calculator.
            self.cosmology = lagos.Cosmology(HubbleConstantNow = 
                                             (100.0 * self.enzoParameters['CosmologyHubbleConstantNow']),
                                             OmegaMatterNow = self.enzoParameters['CosmologyOmegaMatterNow'],
                                             OmegaLambdaNow = self.enzoParameters['CosmologyOmegaLambdaNow'])

            # Instantiate EnzoCosmology object for units and time conversions.
            self.enzo_cosmology = lagos.EnzoCosmology(HubbleConstantNow = 
                                                 (100.0 * self.enzoParameters['CosmologyHubbleConstantNow']),
                                                 OmegaMatterNow = self.enzoParameters['CosmologyOmegaMatterNow'],
                                                 OmegaLambdaNow = self.enzoParameters['CosmologyOmegaLambdaNow'],
                                                 InitialRedshift = self.enzoParameters['CosmologyInitialRedshift'])
            if self.InitialRedshift is not None:
                self.InitialTime = self.enzo_cosmology.ComputeTimeFromRedshift(self.InitialRedshift) / \
                    self.enzo_cosmology.TimeUnits
            if self.FinalRedshift is not None:
                self.FinalTime = self.enzo_cosmology.ComputeTimeFromRedshift(self.FinalRedshift) / \
                    self.enzo_cosmology.TimeUnits

        # Get initial time of simulation.
        if self.enzoParameters['ComovingCoordinates'] and \
                'CosmologyInitialRedshift' in self.enzoParameters:
            self.SimulationInitialTime = self.enzo_cosmology.InitialTime / self.enzo_cosmology.TimeUnits
        elif 'InitialTime' in self.enzoParameters:
            self.SimulationInitialTime = self.enzoParameters['InitialTime']
        else:
            self.SimulationInitialTime = 0.0

        # Combine all data dumps.
        self._CombineDataOutputs()

    def _CalculateRedshiftDumpTimes(self):
        "Calculates time from redshift of redshift dumps."

        for output in self.redshiftOutputs:
            output['time'] = self.enzo_cosmology.ComputeTimeFromRedshift(output['redshift']) / \
                self.enzo_cosmology.TimeUnits

    def _CalculateTimeDumps(self):
        "Calculates time dumps and their redshifts if cosmological."

        if self.enzoParameters['dtDataDump'] <= 0.0: return

        index = 0
        current_time = self.SimulationInitialTime
        while (current_time <= self.FinalTime) or \
                (abs(self.FinalTime - current_time) / self.FinalTime) < dt_Tolerance:
            filename = "%s/%s%04d/%s%04d" % (self.enzoParameters['GlobalDir'],
                                             self.enzoParameters['DataDumpDir'],index,
                                             self.enzoParameters['DataDumpName'],index)

            output = {'index':index,'filename':filename,'time':current_time}
            if self.enzoParameters['ComovingCoordinates']:
                t = self.enzo_cosmology.InitialTime + (self.enzoParameters['dtDataDump'] * self.enzo_cosmology.TimeUnits * index)
                output['redshift'] = self.enzo_cosmology.ComputeRedshiftFromTime(t)

            if not self.get_available_data or os.path.exists(filename):
                self.timeOutputs.append(output)

            current_time += self.enzoParameters['dtDataDump']
            index += 1

    def _CombineDataOutputs(self):
        "Combines redshift and time data into one sorted list."

        # Calculate redshifts for dt data dumps.
        if self.enzoParameters.has_key('dtDataDump') and self.get_time_outputs:
            self._CalculateTimeDumps()

        # Calculate times for redshift dumps.
        if self.enzoParameters['ComovingCoordinates'] and self.get_redshift_outputs:
            self._CalculateRedshiftDumpTimes()
        else:
            self.redshiftOutputs = []

        if self.get_time_outputs and self.get_redshift_outputs:
            self.allOutputs = self.redshiftOutputs + self.timeOutputs
        elif self.get_time_outputs and not self.get_redshift_outputs:
            self.allOutputs = self.timeOutputs
        elif not self.get_time_outputs and self.get_redshift_outputs:
            self.allOutputs = self.redshiftOutputs

        self.allOutputs.sort(key=lambda obj:obj['time'])

        start_index = None
        end_index = None

        for q in range(len(self.allOutputs)):
            del self.allOutputs[q]['index']

            if start_index is None:
                if self.allOutputs[q]['time'] >= self.InitialTime or \
                        abs(self.allOutputs[q]['time'] - self.InitialTime) < \
                        self.allOutputs[q]['time'] * dt_Tolerance:
                    start_index = q

            if end_index is None:
                if self.allOutputs[q]['time'] > self.FinalTime:
                    end_index = q - 1
                if abs(self.allOutputs[q]['time'] - self.FinalTime) < \
                        self.allOutputs[q]['time'] * dt_Tolerance:
                    end_index = q
                if q == len(self.allOutputs) - 1:
                    end_index = q

        for q in range(len(self.allOutputs)):
            if self.links and start_index is not None:
                if q <= start_index:
                    self.allOutputs[q]['previous'] = None
                else:
                    self.allOutputs[q]['previous'] = self.allOutputs[q-1]

                if q >= end_index:
                    self.allOutputs[q]['next'] = None
                else:
                    self.allOutputs[q]['next'] = self.allOutputs[q+1]

        del self.redshiftOutputs
        del self.timeOutputs

        if end_index is None:
            end_index = len(self.allOutputs)-1

        self.allOutputs = self.allOutputs[start_index:end_index+1]

    def _ReadEnzoParameterFile(self):
        "Reads an Enzo parameter file looking for cosmology and output parameters."
        lines = open(self.EnzoParameterFile).readlines()
        for line in lines:
            if line.find("#") >= 0: # Keep the commented lines
                line=line[:line.find("#")]
            line=line.strip()
            if len(line) < 2:
                continue
            try:
                param, vals = map(str,line.split("="))
                param = param.strip()
                vals = vals.strip().split("#", 1)[0].split("//", 1)[0]
            except ValueError:
                continue
            if EnzoParameterDict.has_key(param):
                t = map(EnzoParameterDict[param], vals.split())
                if len(t) == 1:
                    self.enzoParameters[param] = t[0]
                else:
                    self.enzoParameters[param] = t
            elif param.startswith("CosmologyOutputRedshift["):
                index = param[param.find("[")+1:param.find("]")]
                self.redshiftOutputs.append({'index':int(index), 'redshift':float(vals)})

        # Add filenames to redshift outputs.
        tempRedshiftList = []
        for output in self.redshiftOutputs:
            output["filename"] = "%s/%s%04d/%s%04d" % (self.enzoParameters['GlobalDir'],
                                                       self.enzoParameters['RedshiftDumpDir'],output['index'],
                                                       self.enzoParameters['RedshiftDumpName'],output['index'])
            if not self.get_available_data or os.path.exists(output["filename"]):
                tempRedshiftList.append(output)
        self.redshiftOutputs = tempRedshiftList
        del tempRedshiftList

    def _SetParameterDefaults(self):
        "Set some default parameters to avoid problems if they are not in the parameter file."
        self.enzoParameters['GlobalDir'] = "."
        self.enzoParameters['RedshiftDumpName'] = "RD"
        self.enzoParameters['RedshiftDumpDir'] = "RD"
        self.enzoParameters['DataDumpName'] = "data"
        self.enzoParameters['DataDumpDir'] = "DD"
        self.enzoParameters['ComovingCoordinates'] = 0

    def imagine_minimal_splice(self, initial_redshift, final_redshift, decimals=3, filename=None, 
                               redshift_output_string='CosmologyOutputRedshift', start_index=0):
        """
        Create imaginary list of redshift outputs to maximally span a redshift interval.
        :param decimals (int): The decimal place to which the output redshift will be rounded.  
               If the decimal place in question is nonzero, the redshift will be rounded up to 
               ensure continuity of the splice.  Default: 3.
        :param filename (str): If provided, a file will be written with the redshift outputs in 
               the form in which they should be given in the enzo parameter file.  Default: None.
        :param redshift_output_string (str): The parameter accompanying the redshift outputs in the 
               enzo parameter file.  Default: "CosmologyOutputRedshift".
        :param start_index (int): The index of the first redshift output.  Default: 0.
        """

        z = initial_redshift
        outputs = []

        while z > final_redshift:
            rounded = na.round(z, decimals=decimals)
            if rounded - z < 0:
                rounded += na.power(10.0,(-1.0*decimals))
            z = rounded

            deltaz_max = deltaz_forward(self.cosmology, z, self.enzoParameters['CosmologyComovingBoxSize'])
            outputs.append({'redshift': z, 'deltazMax': deltaz_max})
            z -= deltaz_max

        mylog.info("imagine_maximal_splice: Needed %d data dumps to get from z = %f to %f." %
                   (len(outputs), initial_redshift, final_redshift))

        if filename is not None:
            mylog.info("Writing redshift dump list to %s." % filename)
            f = open(filename,'w')
            for q, output in enumerate(outputs):
                z_string = "%%s[%%d] = %%.%df" % decimals
                f.write(("%s[%d] = %." + str(decimals) + "f\n") % (redshift_output_string, (q+start_index), output['redshift']))
            f.close()

        return outputs

    def create_cosmology_splice(self, minimal=True, deltaz_min=0.0, initial_redshift=None, final_redshift=None):
        """
        Create list of datasets to be used for LightCones or LightRays.
        :param minimal (bool): if True, the minimum number of datasets is used to connect the initial and final 
               redshift.  If false, the list will contain as many entries as possible within the redshift 
               interval.  Default: True.
        :param deltaz_min (float): specifies the minimum delta z between consecutive datasets in the returned 
               list.  Default: 0.0.
        :param initial_redshift (float): the initial (highest) redshift in the cosmology splice list.  If none 
               given, the highest redshift dataset present will be used.  Default: None.
        :param final_redshift (float): the final (lowest) redshift in the cosmology splice list.  If none given, 
               the lowest redshift dataset present will be used.  Default: None.
        """

        if initial_redshift is None: initial_redshift = self.InitialRedshift
        if final_redshift is None: final_redshift = self.FinalRedshift

        # Calculate maximum delta z for each data dump.
        self._calculate_deltaz_max()

        # Calculate minimum delta z for each data dump.
        self._calculate_deltaz_min(deltaz_min=deltaz_min)

        cosmology_splice = []

        # Use minimum number of datasets to go from z_i to z_f.
        if minimal:

            z_Tolerance = 1e-3
            z = initial_redshift

            # fill redshift space with datasets
            while ((z > final_redshift) and 
                   (na.fabs(z - final_redshift) > z_Tolerance)):

                # For first data dump, choose closest to desired redshift.
                if (len(cosmology_splice) == 0):
                    # Sort data outputs by proximity to current redsfhit.
                    self.allOutputs.sort(key=lambda obj:na.fabs(z - obj['redshift']))
                    cosmology_splice.append(self.allOutputs[0])

                # Move forward from last slice in stack until z > z_max.
                else:
                    current_slice = cosmology_splice[-1]
                    while current_slice['next'] is not None and \
                            (z < current_slice['next']['redshift'] or \
                                 na.abs(z - current_slice['next']['redshift']) < z_Tolerance):
                        current_slice = current_slice['next']

                    if current_slice is cosmology_splice[-1]:
                        final_redshift = cosmology_splice[-1]['redshift'] - cosmology_splice[-1]['deltazMax']
                        mylog.error("Cosmology splice incomplete due to insufficient data outputs.")
                        break
                    else:
                        cosmology_splice.append(current_slice)

                z = cosmology_splice[-1]['redshift'] - cosmology_splice[-1]['deltazMax']

        # Make light ray using maximum number of datasets (minimum spacing).
        else:
            # Sort data outputs by proximity to current redsfhit.
            self.allOutputs.sort(key=lambda obj:na.fabs(initial_redshift - obj['redshift']))
            # For first data dump, choose closest to desired redshift.
            cosmology_splice.append(self.allOutputs[0])

            nextOutput = cosmology_splice[-1]['next']
            while (nextOutput is not None):
                if (nextOutput['redshift'] <= final_redshift):
                    break
                if ((cosmology_splice[-1]['redshift'] - nextOutput['redshift']) > cosmology_splice[-1]['deltazMin']):
                    cosmology_splice.append(nextOutput)
                nextOutput = nextOutput['next']
            if (cosmology_splice[-1]['redshift'] - cosmology_splice[-1]['deltazMax']) > final_redshift:
                mylog.error("Cosmology splice incomplete due to insufficient data outputs.")
                final_redshift = cosmology_splice[-1]['redshift'] - cosmology_splice[-1]['deltazMax']

        mylog.info("create_cosmology_splice: Used %d data dumps to get from z = %f to %f." % 
                   (len(cosmology_splice),initial_redshift,final_redshift))

        self.allOutputs.sort(key=lambda obj: obj['time'])
        return cosmology_splice

    def get_data_by_redshift(self, redshifts, tolerance=None):
        """
        : param redshifts: a list of redshifts.
        : tolerance: if not None, do not return a dataset unless the redshift is within the tolerance value.
        Get datasets for a list of redshifts.
        """

        redshifts = funcs.ensure_list(redshifts)
        my_datasets = []
        for redshift in redshifts:
            self.allOutputs.sort(key=lambda obj:na.fabs(redshift - obj['redshift']))
            if (tolerance is None or na.abs(redshift - self.allOutputs[0]['redshift']) <= tolerance) \
                    and self.allOutputs[0] not in my_datasets:
                my_datasets.append(self.allOutputs[0])
            else:
                mylog.error("No dataset added for z = %f." % redshift)

        self.allOutputs.sort(key=lambda obj: obj['time'])
        return my_datasets

    def get_data_by_time(self, times, tolerance=None):
        """
        : param redshifts: a list of times.
        : tolerance: if not None, do not return a dataset unless the redshift is within the tolerance value.
        Get datasets for a list of times.
        """

        times = funcs.ensure_list(times)
        my_datasets = []
        for my_time in times:
            self.allOutputs.sort(key=lambda obj:na.fabs(my_time - obj['time']))
            if (tolerance is None or na.abs(my_time - self.allOutputs[0]['time']) <= tolerance) \
                    and self.allOutputs[0] not in my_datasets:
                my_datasets.append(self.allOutputs[0])
            else:
                mylog.error("No dataset added for z = %f." % my_time)

        self.allOutputs.sort(key=lambda obj: obj['time'])
        return my_datasets

    def _calculate_deltaz_max(self):
        "Calculate delta z that corresponds to full box length going from z to (z - delta z)."

        d_Tolerance = 1e-4
        max_Iterations = 100

        targetDistance = self.enzoParameters['CosmologyComovingBoxSize']

        for output in self.allOutputs:
            z = output['redshift']

            # Calculate delta z that corresponds to the length of the box at a given redshift.
            # Use Newton's method to calculate solution.
            z1 = z
            z2 = z1 - 0.1 # just an initial guess
            distance1 = 0.0
            iteration = 1

            # Convert comoving radial distance into Mpc / h, since that's how box size is stored.
            distance2 = self.cosmology.ComovingRadialDistance(z2,z) * self.enzoParameters['CosmologyHubbleConstantNow']

            while ((na.fabs(distance2-targetDistance)/distance2) > d_Tolerance):
                m = (distance2 - distance1) / (z2 - z1)
                z1 = z2
                distance1 = distance2
                z2 = ((targetDistance - distance2) / m) + z2
                distance2 = self.cosmology.ComovingRadialDistance(z2,z) * self.enzoParameters['CosmologyHubbleConstantNow']
                iteration += 1
                if (iteration > max_Iterations):
                    mylog.error("calculate_deltaz_max: Warning - max iterations exceeded for z = %f (delta z = %f)." % 
                                (z,na.fabs(z2-z)))
                    break
            output['deltazMax'] = na.fabs(z2-z)

    def _calculate_deltaz_min(self, deltaz_min=0.0):
        "Calculate delta z that corresponds to a single top grid pixel going from z to (z - delta z)."

        d_Tolerance = 1e-4
        max_Iterations = 100

        targetDistance = self.enzoParameters['CosmologyComovingBoxSize'] / self.enzoParameters['TopGridDimensions'][0]

        for output in self.allOutputs:
            z = output['redshift']

            # Calculate delta z that corresponds to the length of a top grid pixel at a given redshift.
            # Use Newton's method to calculate solution.
            z1 = z
            z2 = z1 - 0.01 # just an initial guess
            distance1 = 0.0
            iteration = 1

            # Convert comoving radial distance into Mpc / h, since that's how box size is stored.
            distance2 = self.cosmology.ComovingRadialDistance(z2,z) * self.enzoParameters['CosmologyHubbleConstantNow']

            while ((na.fabs(distance2 - targetDistance) / distance2) > d_Tolerance):
                m = (distance2 - distance1) / (z2 - z1)
                z1 = z2
                distance1 = distance2
                z2 = ((targetDistance - distance2) / m) + z2
                distance2 = self.cosmology.ComovingRadialDistance(z2,z) * self.enzoParameters['CosmologyHubbleConstantNow']
                iteration += 1
                if (iteration > max_Iterations):
                    mylog.error("calculate_deltaz_max: Warning - max iterations exceeded for z = %f (delta z = %f)." % 
                                (z,na.fabs(z2-z)))
                    break
            # Use this calculation or the absolute minimum specified by the user.
            output['deltazMin'] = max(na.fabs(z2-z),deltaz_min)

EnzoParameterDict = {"CosmologyOmegaMatterNow": float,
                     "CosmologyOmegaLambdaNow": float,
                     "CosmologyHubbleConstantNow": float,
                     "CosmologyInitialRedshift": float,
                     "CosmologyFinalRedshift": float,
                     "ComovingCoordinates": int,
                     "InitialTime": float,
                     "StopTime": float,
                     "TopGridDimensions": float,
                     "dtDataDump": float,
                     "RedshiftDumpName": str,
                     "RedshiftDumpDir":  str,
                     "DataDumpName": str,
                     "DataDumpDir": str,
                     "GlobalDir" : str}

def deltaz_forward(cosmology, z, target_distance):
    "Calculate deltaz corresponding to moving a comoving distance starting from some redshift."

    d_Tolerance = 1e-4
    max_Iterations = 100

    # Calculate delta z that corresponds to the length of the box at a given redshift.
    # Use Newton's method to calculate solution.
    z1 = z
    z2 = z1 - 0.1 # just an initial guess
    distance1 = 0.0
    iteration = 1

    # Convert comoving radial distance into Mpc / h, since that's how box size is stored.
    distance2 = cosmology.ComovingRadialDistance(z2,z) * cosmology.HubbleConstantNow / 100.0

    while ((na.fabs(distance2-target_distance)/distance2) > d_Tolerance):
        m = (distance2 - distance1) / (z2 - z1)
        z1 = z2
        distance1 = distance2
        z2 = ((target_distance - distance2) / m) + z2
        distance2 = cosmology.ComovingRadialDistance(z2,z) * cosmology.HubbleConstantNow / 100.0
        iteration += 1
        if (iteration > max_Iterations):
            mylog.error("deltaz_forward: Warning - max iterations exceeded for z = %f (delta z = %f)." % 
                        (z,na.fabs(z2-z)))
            break
    return na.fabs(z2-z)



"""
Simple readers for fortran unformatted data, specifically for the Tiger code.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2010 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython

from stdio cimport fopen, fclose, FILE

cdef extern from "stdio.h":
    cdef int SEEK_SET
    cdef int SEEK_CUR
    cdef int SEEK_END
    int fseek(FILE *stream, long offset, int whence)
    size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream)
    long ftell(FILE *stream)

@cython.boundscheck(False)
@cython.wraparound(False)
def read_tiger_section(
                     char *fn,
                     np.ndarray[np.int64_t, ndim=1] slab_start,
                     np.ndarray[np.int64_t, ndim=1] slab_size,
                     np.ndarray[np.int64_t, ndim=1] root_size,
                     int offset = 36):
    cdef int strides[3]
    strides[0] = 1
    strides[1] = root_size[0] * strides[0]
    strides[2] = strides[1] * root_size[1] + 2
    cdef np.int64_t i, j, k
    cdef np.ndarray buffer = np.zeros(slab_size, dtype='float32', order='F')
    cdef FILE *f = fopen(fn, "rb")
    #for i in range(3): offset += strides[i] * slab_start[i]
    cdef np.int64_t pos = 0
    cdef np.int64_t moff = 0
    cdef float *data = <float *> buffer.data
    fseek(f, offset, 0)
    # If anybody wants to convert this loop to a SEEK_CUR, that'd be great.
    for i in range(slab_size[2]):
        for j in range(slab_size[1]):
            moff = (slab_start[0]    ) * strides[0] \
                 + (slab_start[1] + j) * strides[1] \
                 + (slab_start[2] + i) * strides[2]
            #print offset + 4 * moff, pos
            fseek(f, offset + 4 * moff, SEEK_SET)
            fread(<void *> (data + pos), 4, slab_size[0], f)
            pos += slab_size[0]
    return buffer



"""
Halo filters to be used with the HaloProfiler.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.logger import lagosLogger as mylog
from copy import deepcopy
import numpy as na

def VirialFilter(profile,overdensity_field='ActualOverdensity',
                 virial_overdensity=200.,must_be_virialized=True,
                 virial_filters=[['TotalMassMsun','>=','1e14']],
                 virial_quantities=['TotalMassMsun','RadiusMpc'],
                 virial_index=None):
    """
    Filter halos by virial quantities.
    Return values are a True or False whether the halo passed the filter, 
    along with a dictionary of virial quantities for the fields specified in 
    the virial_quantities keyword.  Thresholds for virial quantities are 
    given with the virial_filters keyword in the following way: 
    [field, condition, value].
    """

    fields = deepcopy(virial_quantities)
    if virial_filters is None: virial_filters = []
    for vfilter in virial_filters:
        if not vfilter[0] in fields:
            fields.append(vfilter[0])
    
    overDensity = []
    temp_profile = {}
    for field in fields:
        temp_profile[field] = []

    for q in range(len(profile[overdensity_field])):
        good = True
        if (profile[overdensity_field][q] != profile[overdensity_field][q]):
            good = False
            continue
        for field in fields:
            if (profile[field][q] != profile[field][q]):
                good = False
                break
        if good:
            overDensity.append(profile[overdensity_field][q])
            for field in fields:
                temp_profile[field].append(profile[field][q])

    virial = {}
    for field in fields:
        virial[field] = 0.0

    if (not (na.array(overDensity) >= virial_overdensity).any()) and \
            must_be_virialized:
        mylog.error("This halo is not virialized!")
        return [False, {}]

    if (len(overDensity) < 2):
        mylog.error("Skipping halo with no valid points in profile.")
        return [False, {}]

    if (overDensity[1] <= virial_overdensity):
        index = 0
    elif (overDensity[-1] >= virial_overdensity):
        index = -2
    else:
        for q in (na.arange(len(overDensity)-2))+2:
            if (overDensity[q] < virial_overdensity):
                index = q - 1
                break

    if type(virial_index) is list:
        virial_index.append(index)

    for field in fields:
        if (overDensity[index+1] - overDensity[index]) == 0:
            mylog.error("Overdensity profile has slope of zero.")
            return [False, {}]
        else:
            slope = (temp_profile[field][index+1] - temp_profile[field][index]) / \
                (overDensity[index+1] - overDensity[index])
            value = slope * (virial_overdensity - overDensity[index]) + \
                temp_profile[field][index]
            virial[field] = value

    for vfilter in virial_filters:
        if eval("%s %s %s" % (virial[vfilter[0]],vfilter[1],vfilter[2])):
            mylog.debug("(%s %s %s) returned True for %s." % (vfilter[0],vfilter[1],vfilter[2],virial[vfilter[0]]))
            continue
        else:
            mylog.debug("(%s %s %s) returned False for %s." % (vfilter[0],vfilter[1],vfilter[2],virial[vfilter[0]]))
            return [False, {}]

    return [True, dict((q,virial[q]) for q in virial_quantities)]




"""
Import stuff for light cone generator.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import yt.lagos as lagos

from LightConeProjection import *
from light_cone import *
from HaloMask import *
from UniqueSolution import *



"""
Various definitions for various other modules and routines

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/

@todo: Move into yt.Defs, along with enki.EnkiDefs
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# The number of levels we expect to have at most
MAXLEVEL=48

# Number of levels to toss back to check for maximum
NUMTOCHECK=2

axis_labels = [('y','z'),('x','z'),('x','y')]
axis_names = {0: 'x', 1: 'y', 2: 'z', 4:''}
inv_axis_names = {'x':0,'y':1,'z':2}

vm_axis_names = {0:'x', 1:'y', 2:'z', 3:'dx', 4:'dy'}

# The appropriate axes for which way we are slicing
x_dict = [1,0,0]
y_dict = [2,2,1]

x_names = ['y','x','x']
y_names = ['z','z','y']

mh = 1.67e-24
mu = 1.22

# All the parameters we read from the parameter file, along with how to convert
# them from a string
parameterDict = {"CosmologyCurrentRedshift": float,
                 "CosmologyComovingBoxSize": float,
                 "CosmologyOmegaMatterNow": float,
                 "CosmologyOmegaLambdaNow": float,
                 "CosmologyHubbleConstantNow": float,
                 "CosmologyInitialRedshift": float,
                 "DualEnergyFormalismEta1": float,
                 "DualEnergyFormalismEta2": float,
                 "MetaDataString": str,
                 "HydroMethod": int,
                 "DualEnergyFormalism": int,
                 "InitialTime": float,
                 "ComovingCoordinates": int,
                 "DensityUnits": float,
                 "LengthUnits": float,
                 "LengthUnit": float,
                 "TemperatureUnits": float,
                 "TimeUnits": float,
                 "GravitationalConstant": float,
                 "Gamma": float,
                 "MultiSpecies": int,
                 "CompilerPrecision": str,
                 "CurrentTimeIdentifier": int,
                 "RefineBy": int,
                 "BoundaryConditionName": str,
                 "TopGridRank": int,
                 "TopGridDimensions": int,
                 "EOSSoundSpeed": float,
                 "EOSType": int,
                 "NumberOfParticleAttributes": int,
                }

mpc_conversion = {'mpc'   : 1e0,
                  'kpc'   : 1e3,
                  'pc'    : 1e6,
                  'au'    : 2.063e11,
                  'rsun'  : 4.43664e13,
                  'cm'    : 3.0857e24,
                  'miles' : 1.917e19}

axis_labels = [('y','z'),('x','z'),('x','y')]



#!/usr/bin/env python
import setuptools
import os, sys

def check_for_hdf5():
    # First up: HDF5_DIR in environment
    if "HDF5_DIR" in os.environ:
        hdf5_dir = os.environ["HDF5_DIR"]
        hdf5_inc = os.path.join(hdf5_dir, "include")
        hdf5_lib = os.path.join(hdf5_dir, "lib")
        print "HDF5_LOCATION: HDF5_DIR: %s, %s" % (hdf5_inc, hdf5_lib)
        return (hdf5_inc, hdf5_lib)
    # Next up, we try hdf5.cfg
    elif os.path.exists("hdf5.cfg"):
        hdf5_dir = open("hdf5.cfg").read().strip()
        hdf5_inc = os.path.join(hdf5_dir, "include")
        hdf5_lib = os.path.join(hdf5_dir, "lib")
        print "HDF5_LOCATION: hdf5.cfg: %s, %s" % (hdf5_inc, hdf5_lib)
        return (hdf5_inc, hdf5_lib)
    # Now we see if ctypes can help us:
    try:
        import ctypes.util
        hdf5_libfile = ctypes.util.find_library("hdf5")
        if hdf5_libfile is not None and os.path.isfile(hdf5_libfile):
            # Now we've gotten a library, but we'll need to figure out the
            # includes if this is going to work.  It feels like there is a
            # better way to pull off two directory names.
            hdf5_dir = os.path.dirname(os.path.dirname(hdf5_libfile))
            if os.path.isdir(os.path.join(hdf5_dir, "include")) and \
               os.path.isfile(os.path.join(hdf5_dir, "include", "hdf5.h")):
                hdf5_inc = os.path.join(hdf5_dir, "include")
                hdf5_lib = os.path.join(hdf5_dir, "lib")
                print "HDF5_LOCATION: HDF5 found in: %s, %s" % (hdf5_inc, hdf5_lib)
                return hdf5_inc, hdf5_lib
    except ImportError:
        pass
    print "Reading HDF5 location from hdf5.cfg failed."
    print "Please place the base directory of your HDF5 install in hdf5.cfg and restart."
    print "(ex: \"echo '/usr/local/' > hdf5.cfg\" )"
    sys.exit(1)

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('lagos',parent_package,top_path)
    config.make_config_py() # installs __config__.py
    config.make_svn_version_py()
    config.add_extension("PointCombine", "yt/lagos/PointCombine.c", libraries=["m"])
    config.add_subpackage("hop")
    config.add_subpackage("fof")
    config.add_subpackage("parallelHOP")
    hdf5_inc, hdf5_lib = check_for_hdf5()
    include_dirs=[hdf5_inc]
    library_dirs=[hdf5_lib]
    config.add_extension("HDF5LightReader", "yt/lagos/HDF5LightReader.c",
                         define_macros=[("H5_USE_16_API",True)],
                         libraries=["m","hdf5"],
                         library_dirs=library_dirs, include_dirs=include_dirs)
    return config



"""
Very simple convenience function for importing all the extension modules.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

#
# ALL IMPORTS GO HERE
#

# First module imports
import halo_profiler as HP
import lightcone as LC
import SpectralIntegrator as SI
import RayTracer as RT
import EnzoSimulation as ES

import HierarchySubset
import disk_analysis
import coordinate_transforms

# Now individual component imports
from halo_profiler import HaloProfiler
from lightcone import LightCone
from SpectralIntegrator import create_table_from_textfiles, \
                               SpectralFrequencyIntegrator
from RayTracer import SlowRayTracer
from HierarchySubset import ExtractedHierarchy

from disk_analysis import StackedDiskImage
from coordinate_transforms import spherical_regrid, arbitrary_regrid



"""
Output-watcher

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.fido import *

class Watcher:
    def __init__(self, title=None, path=".", new_prefix="", oc=None,
                 process=None, function_handler = None):
        self.original_path = os.getcwd()
        os.chdir(path)
        if title == None: title = os.path.basename(os.getcwd())
        if oc == None: oc = OutputCollection(title)
        self.title = title
        self.oc = oc
        self.process = process
        self.new_prefix = new_prefix
        self.skip_files = [] # Forward compatible
        self.function_handler = None
        if function_handler != None:
            self.function_handler = function_handler()

    def run(self, run_once=False):
        mylog.info("Entering main Fido loop (CTRL-C or touch 'stopFido' to end)")
        wb = ytcfg.getfloat("fido","WaitBetween")
        while not self.check_for_stop():
            nn = self.check_for_output()
            for bn in sorted(nn):
                new_name = bury_output(bn, new_prefix=self.new_prefix)
                self.handle_output(new_name)
            if run_once: break
            try:
                time.sleep(wb)
            except KeyboardInterrupt:
                sys.exit()

    def handle_output(self, filename):
        # First, add it to the OutputCollection
        self.oc.add_output(filename)
        self.oc.write_out()
        if self.function_handler == None: return
        # Now, we pass it to our function handler
        pid = os.fork()
        if pid:
            mylog.debug("Waiting on pid %s", pid)
            newpid, exit = os.waitpid(pid,0)
            mylog.debug("Exit status %s from PID %s", exit, newpid)
        else:
            mylog.info("Forked process reporting for duty!")
            self.function_handler(filename)
            sys.exit()

    def check_for_output(self):
        new_files = []
        if os.path.isfile(NEW_OUTPUT_CREATED):
            os.unlink(NEW_OUTPUT_CREATED)
            # So something is created!  Now let's snag it
            # We insert our additional glob patterns here
            files_found = glob.glob("*.hierarchy") #\
                         #[glob.glob(pat) for pat in GlobPatterns]
            for file in [fn for fn in files_found if fn not in self.skip_files]:
                #if self.oc.has_key(os.path.basename(file[:-10])): continue
                new_files.append(file.rsplit(".",1)[0])
                mylog.info("Found output %s", new_files[-1])
        return new_files

    def check_for_stop(self):
        if os.path.exists("stopFido"):
            # We should log this rather than print it
            mylog.info("Stopping fido")
            os.unlink("stopFido")
            return 1
        if self.process:
            pp = self.process.poll()
            if pp != None:
                mylog.info("Process has died; stopping fido")
                return 1
        return 0




"""
This is a definition of a class for defining problem verification.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.mods import *
from yt.funcs import *
from yt.extensions.EnzoSimulation import *
from SimulationTests import test_registry

from mercurial import commands, hg, ui
import cPickle

class Problem(object):
    def __init__(self, name, pfname, repo_path="."):
        self.name = name
        self.pfname = pfname
        self.simulation = EnzoSimulation(pfname)
        self.tests, self.test_types = [], {}
        self.init_repo(repo_path)
        self._initialize_tests()

    def _add_existing_test(self, test_type):
        @wraps(test_type.__init__)
        def test_wrapper(name, *args, **kwargs):
            new_test = test_type(name, *args, **kwargs)
            self.tests.append(new_test)
        return test_wrapper

    def _initialize_tests(self):
        for name, test in test_registry.iteritems():
            setattr(self, 'add_%s' % name, self._add_existing_test(test))
            self.test_types[name] = self._add_existing_test(test)

    def init_repo(self, path):
        self.ui = ui.ui()
        try:
            self.repo = hg.repository(self.ui, path)
        except:
            print "CREATING:", path
            self.repo = hg.repository(self.ui, path, create=True)

    def load_repo_file(self, fn, identifier='tip'):
        return self.repo[identifier][fn].data()

    def run_tests(self):
        all_results = {}
        try:
            for pf in self.simulation:
                results = {}
                for test in self.tests:
                    results[test.name] = test(pf)
                all_results[str(pf)] = results
        except IOError:
            pass
        return all_results

    def store_results(self, results):
        base_fn = "results_%s.cpkl" % self.name
        fn = self.repo.pathto(base_fn)
        cPickle.dump(results, open(fn, "w"))
        if base_fn not in self.repo['tip'].manifest():
            commands.add(self.ui, self.repo, fn)
        message = "Committing results from current run"
        commands.commit(self.ui, self.repo, fn, message=message)
        print "Committed"

    def __call__(self):
        results = self.run_tests()
        self.store_results((self.tests, results))

# repo['tip']['yt/lagos/__init__.py'].data()



"""
Fields applicable only to Enzo

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2008-2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from UniversalFields import *
from yt.amr_utils import CICDeposit_3

rho_crit_now = 1.8788e-29 # times h^2

class EnzoFieldContainer(CodeFieldInfoContainer):
    """
    This is a container for Enzo-specific fields.
    """
    _shared_state = {}
    _field_list = {}
EnzoFieldInfo = EnzoFieldContainer()
add_enzo_field = EnzoFieldInfo.add_field

add_field = add_enzo_field

_speciesList = ["HI","HII","Electron",
               "HeI","HeII","HeIII",
               "H2I","H2II","HM",
               "DI","DII","HDI","Metal","PreShock"]
_speciesMass = {"HI":1.0,"HII":1.0,"Electron":1.0,
                "HeI":4.0,"HeII":4.0,"HeIII":4.0,
                "H2I":2.0,"H2II":2.0,"HM":1.0,
                "DI":2.0,"DII":2.0,"HDI":3.0}

def _SpeciesComovingDensity(field, data):
    sp = field.name.split("_")[0] + "_Density"
    ef = (1.0 + data.pf["CosmologyCurrentRedshift"])**3.0
    return data[sp]/ef
def _SpeciesFraction(field, data):
    sp = field.name.split("_")[0] + "_Density"
    return data[sp]/data["Density"]
def _SpeciesNumberDensity(field, data):
    species = field.name.split("_")[0]
    sp = field.name.split("_")[0] + "_Density"
    return data[sp]/_speciesMass[species]
def _ConvertNumberDensity(data):
    return 1.0/mh

for species in _speciesList:
    add_field("%s_Fraction" % species,
             function=_SpeciesFraction,
             validators=ValidateDataField("%s_Density" % species))
    add_field("Comoving_%s_Density" % species,
             function=_SpeciesComovingDensity,
             validators=ValidateDataField("%s_Density" % species))
    if _speciesMass.has_key(species):
        add_field("%s_NumberDensity" % species,
                  function=_SpeciesNumberDensity,
                  convert_function=_ConvertNumberDensity,
                  validators=ValidateDataField("%s_Density" % species))

def _Metallicity(field, data):
    return data["Metal_Fraction"]
def _ConvertMetallicity(data):
    return 49.0196 # 1 / 0.0204
add_field("Metallicity", units=r"Z_{\rm{\odot}}",
          function=_Metallicity,
          convert_function=_ConvertMetallicity,
          validators=ValidateDataField("Metal_Density"),
          projection_conversion="1")

def _Metallicity3(field, data):
    return data["SN_Colour"]/data["Density"]
add_field("Metallicity3", units=r"Z_{\rm{\odot}}",
          function=_Metallicity3,
          convert_function=_ConvertMetallicity,
          validators=ValidateDataField("SN_Colour"),
          projection_conversion="1")

def _Cooling_Time(field, data):
    return data["Cooling_Time"]
add_field("Cooling_Time", units=r"\rm{s}",
          function=_Cooling_Time,
          validators=ValidateDataField("Cooling_Time"),
          projection_conversion="1")

def _ThermalEnergy(field, data):
    if data.pf["HydroMethod"] == 2:
        return data["Total_Energy"]
    else:
        if data.pf["DualEnergyFormalism"]:
            return data["GasEnergy"]
        else:
            return data["Total_Energy"] - 0.5*(
                   data["x-velocity"]**2.0
                 + data["y-velocity"]**2.0
                 + data["z-velocity"]**2.0 )
add_field("ThermalEnergy", function=_ThermalEnergy,
          units=r"\rm{ergs}/\rm{cm^3}")

# This next section is the energy field section
# Note that we have aliases that manually unconvert themselves.
# This is because numerous code branches use Gas_Energy or GasEnergy
# indiscriminately -- this is almost fixed with LCA1.5, but not everyone is
# moving to that branch.  So, because the actual function doesn't get called
# *unless* it's an alias, we simply de-convert -- since the input data is
# already converted to cgs.

def _convertEnergy(data):
    return data.convert("x-velocity")**2.0

def _GasEnergy(field, data):
    return data["Gas_Energy"] / _convertEnergy(data)
add_field("GasEnergy", function=_GasEnergy,
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

def _Gas_Energy(field, data):
    return data["GasEnergy"] / _convertEnergy(data)
add_field("Gas_Energy", function=_Gas_Energy,
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

def _TotalEnergy(field, data):
    return data["Total_Energy"] / _convertEnergy(data)
add_field("TotalEnergy", function=_TotalEnergy,
          display_name = "\mathrm{Total}\/\mathrm{Energy}",
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

def _Total_Energy(field, data):
    return data["TotalEnergy"] / _convertEnergy(data)
add_field("Total_Energy", function=_Total_Energy,
          display_name = "\mathrm{Total}\/\mathrm{Energy}",
          units=r"\rm{ergs}/\rm{g}", convert_function=_convertEnergy)

def _NumberDensity(field, data):
    # We can assume that we at least have Density
    # We should actually be guaranteeing the presence of a .shape attribute,
    # but I am not currently implementing that
    fieldData = na.zeros(data["Density"].shape,
                         dtype = data["Density"].dtype)
    if data.pf["MultiSpecies"] == 0:
        if data.has_field_parameter("mu"):
            mu = data.get_field_parameter("mu")
        else:
            mu = 0.6
        fieldData += data["Density"] / mu
    if data.pf["MultiSpecies"] > 0:
        fieldData += data["HI_Density"] / 1.0
        fieldData += data["HII_Density"] / 1.0
        fieldData += data["HeI_Density"] / 4.0
        fieldData += data["HeII_Density"] / 4.0
        fieldData += data["HeIII_Density"] / 4.0
        fieldData += data["Electron_Density"] / 1.0
    if data.pf["MultiSpecies"] > 1:
        fieldData += data["HM_Density"] / 1.0
        fieldData += data["H2I_Density"] / 2.0
        fieldData += data["H2II_Density"] / 2.0
    if data.pf["MultiSpecies"] > 2:
        fieldData += data["DI_Density"] / 2.0
        fieldData += data["DII_Density"] / 2.0
        fieldData += data["HDI_Density"] / 3.0
    return fieldData
add_field("NumberDensity", units=r"\rm{cm}^{-3}",
          function=_NumberDensity,
          convert_function=_ConvertNumberDensity)

def _ComovingDensity(field,data):
    ef = (1.0 + data.pf["CosmologyCurrentRedshift"])**3.0
    return data["Density"]/ef
add_field("ComovingDensity", function=_ComovingDensity, units=r"\rm{g}/\rm{cm}^3")

def Overdensity(field,data):
    return (data['Density'] + data['Dark_Matter_Density']) / \
        (rho_crit_now * (data.pf['CosmologyHubbleConstantNow']**2) * ((1+data.pf['CosmologyCurrentRedshift'])**3))
add_field("Overdensity",function=Overdensity,units=r"")

# Now we add all the fields that we want to control, but we give a null function
# This is every Enzo field we can think of.  This will be installation-dependent,

# removed: "Gas_Energy","Total_Energy",
# these are now aliases for each other

_default_fields = ["Density","Temperature",
                   "x-velocity","y-velocity","z-velocity",
                   "x-momentum","y-momentum","z-momentum"]
# else:
#     _default_fields = ["Density","Temperature","Gas_Energy","Total_Energy",
#                        "x-velocity","y-velocity","z-velocity"]
_default_fields += [ "%s_Density" % sp for sp in _speciesList ]

for field in _default_fields:
    add_field(field, function=lambda a, b: None, take_log=True,
              validators=[ValidateDataField(field)], units=r"\rm{g}/\rm{cm}^3")
EnzoFieldInfo["x-velocity"].projection_conversion='1'
EnzoFieldInfo["y-velocity"].projection_conversion='1'
EnzoFieldInfo["z-velocity"].projection_conversion='1'

# Now we override

def _convertDensity(data):
    return data.convert("Density")
for field in ["Density"] + [ "%s_Density" % sp for sp in _speciesList ]:
    EnzoFieldInfo[field]._units = r"\rm{g}/\rm{cm}^3"
    EnzoFieldInfo[field]._projected_units = r"\rm{g}/\rm{cm}^2"
    EnzoFieldInfo[field]._convert_function=_convertDensity

add_field("Dark_Matter_Density", function=lambda a,b: None,
          convert_function=_convertDensity,
          validators=[ValidateDataField("Dark_Matter_Density"),
                      ValidateSpatial(0)],
          display_name = "Dark\ Matter\ Density",
          not_in_all = True)

EnzoFieldInfo["Temperature"]._units = r"\rm{K}"
EnzoFieldInfo["Temperature"].units = r"K"

def _convertVelocity(data):
    return data.convert("x-velocity")
for ax in ['x','y','z']:
    f = EnzoFieldInfo["%s-velocity" % ax]
    f._units = r"\rm{cm}/\rm{s}"
    f._convert_function = _convertVelocity
    f.take_log = False

def _pdensity(field, data):
    blank = na.zeros(data.ActiveDimensions, dtype='float32', order="FORTRAN")
    if data.NumberOfParticles == 0: return blank
    cic_deposit.cic_deposit(data["particle_position_x"],
                            data["particle_position_y"],
                            data["particle_position_z"], 3,
                            data["particle_mass"],
                            blank, data.LeftEdge, data['dx'])
    return blank
add_field("particle_density", function=_pdensity,
          validators=[ValidateSpatial(0)], convert_function=_convertDensity)

def _spdensity_pyx(field, data):
    blank = na.zeros(data.ActiveDimensions, dtype='float32')
    if data.NumberOfParticles == 0: return blank
    filter = data['creation_time'] > 0.0
    if not filter.any(): return blank
    CICDeposit_3(data["particle_position_x"][filter].astype(na.float64),
                 data["particle_position_y"][filter].astype(na.float64),
                 data["particle_position_z"][filter].astype(na.float64),
                 data["particle_mass"][filter].astype(na.float32),
                 na.int64(na.where(filter)[0].size),
                 blank, na.array(data.LeftEdge).astype(na.float64),
                 na.array(data.ActiveDimensions).astype(na.int32), 
                 na.float64(data['dx']))
    return blank
add_field("star_density_pyx", function=_spdensity_pyx,
          validators=[ValidateSpatial(0)], convert_function=_convertDensity)

def _dmpdensity_pyx(field, data):
    blank = na.zeros(data.ActiveDimensions, dtype='float32')
    if data.NumberOfParticles == 0: return blank
    filter = data['creation_time'] <= 0.0
    if not filter.any(): return blank
    CICDeposit_3(data["particle_position_x"][filter].astype(na.float64),
                 data["particle_position_y"][filter].astype(na.float64),
                 data["particle_position_z"][filter].astype(na.float64),
                 data["particle_mass"][filter].astype(na.float32),
                 na.int64(na.where(filter)[0].size),
                 blank, na.array(data.LeftEdge).astype(na.float64),
                 na.array(data.ActiveDimensions).astype(na.int32), 
                 na.float64(data['dx']))
    return blank
add_field("dm_density_pyx", function=_dmpdensity_pyx,
          validators=[ValidateSpatial(0)], convert_function=_convertDensity)

def _star_field(field, data):
    """
    Create a grid field for star quantities, weighted by star mass.
    """
    particle_field = field.name[5:]
    top = na.zeros(data.ActiveDimensions, dtype='float32')
    if data.NumberOfParticles == 0: return top
    filter = data['creation_time'] > 0.0
    if not filter.any(): return top
    particle_field_data = data[particle_field][filter] * data['particle_mass'][filter]
    CICDeposit_3(data["particle_position_x"][filter].astype(na.float64),
                 data["particle_position_y"][filter].astype(na.float64),
                 data["particle_position_z"][filter].astype(na.float64),
                 particle_field_data.astype(na.float32),
                 na.int64(na.where(filter)[0].size),
                 top, na.array(data.LeftEdge).astype(na.float64),
                 na.array(data.ActiveDimensions).astype(na.int32), 
                 na.float64(data['dx']))
    del particle_field_data

    bottom = na.zeros(data.ActiveDimensions, dtype='float32')
    CICDeposit_3(data["particle_position_x"][filter].astype(na.float64),
                 data["particle_position_y"][filter].astype(na.float64),
                 data["particle_position_z"][filter].astype(na.float64),
                 data["particle_mass"][filter].astype(na.float32),
                 na.int64(na.where(filter)[0].size),
                 bottom, na.array(data.LeftEdge).astype(na.float64),
                 na.array(data.ActiveDimensions).astype(na.int32), 
                 na.float64(data['dx']))

    top[bottom == 0] = 0.0
    bnz = bottom.nonzero()
    top[bnz] /= bottom[bnz]
    return top

add_field('star_metallicity_fraction', function=_star_field,
          validators=[ValidateSpatial(0)])
add_field('star_creation_time', function=_star_field,
          validators=[ValidateSpatial(0)])
add_field('star_dynamical_time', function=_star_field,
          validators=[ValidateSpatial(0)])

def _StarMetallicity(field, data):
    return data['star_metallicity_fraction']
add_field('StarMetallicity', units=r"Z_{\rm{\odot}}",
          function=_StarMetallicity,
          convert_function=_ConvertMetallicity,
          projection_conversion="1")

def _StarCreationTime(field, data):
    return data['star_creation_time']
def _ConvertEnzoTimeYears(data):
    return data.pf.time_units['years']
add_field('StarCreationTimeYears', units=r"\mathrm{yr}",
          function=_StarCreationTime,
          convert_function=_ConvertEnzoTimeYears,
          projection_conversion="1")

def _StarDynamicalTime(field, data):
    return data['star_dynamical_time']
add_field('StarDynamicalTimeYears', units=r"\mathrm{yr}",
          function=_StarDynamicalTime,
          convert_function=_ConvertEnzoTimeYears,
          projection_conversion="1")

def _StarAge(field, data):
    star_age = na.zeros(data['StarCreationTimeYears'].shape)
    with_stars = data['StarCreationTimeYears'] > 0
    star_age[with_stars] = data.pf.time_units['years'] * \
        data.pf["InitialTime"] - \
        data['StarCreationTimeYears'][with_stars]
    return star_age
add_field('StarAgeYears', units=r"\mathrm{yr}",
          function=_StarAge,
          projection_conversion="1")

def _IsStarParticle(field, data):
    is_star = (data['creation_time'] > 0).astype('float64')
    return is_star
add_field('IsStarParticle', function=_IsStarParticle,
          particle_type = True)

#
# Now we do overrides for 2D fields
#

class Enzo2DFieldContainer(CodeFieldInfoContainer):
    _shared_state = {}
    _field_list = EnzoFieldContainer._field_list.copy()
# We make a copy of the dict from the other, so we
# can now update it...
Enzo2DFieldInfo = Enzo2DFieldContainer()
add_enzo_2d_field = Enzo2DFieldInfo.add_field

def _CellArea(field, data):
    if data['dx'].size == 1:
        try:
            return data['dx']*data['dy']*\
                na.ones(data.ActiveDimensions, dtype='float64')
        except AttributeError:
            return data['dx']*data['dy']
    return data["dx"]*data["dy"]
def _ConvertCellAreaMpc(data):
    return data.convert("mpc")**2.0
def _ConvertCellAreaCGS(data):
    return data.convert("cm")**2.0
add_enzo_2d_field("CellAreaCode", units=r"\rm{BoxArea}^2",
          function=_CellArea)
add_enzo_2d_field("CellAreaMpc", units=r"\rm{Mpc}^2",
          function=_CellArea,
          convert_function=_ConvertCellAreaMpc)
add_enzo_2d_field("CellArea", units=r"\rm{cm}^2",
          function=_CellArea,
          convert_function=_ConvertCellAreaCGS)

for a in ["Code", "Mpc", ""]:
    Enzo2DFieldInfo["CellVolume%s" % a] = \
        Enzo2DFieldInfo["CellArea%s" % a]

def _zvel(field, data):
    return na.zeros(data["x-velocity"].shape,
                    dtype='float64')
add_enzo_2d_field("z-velocity", function=_zvel)


#
# Now we do overrides for 1D fields
#

class Enzo1DFieldContainer(CodeFieldInfoContainer):
    _shared_state = {}
    _field_list = EnzoFieldContainer._field_list.copy()
# We make a copy of the dict from the other, so we
# can now update it...
Enzo1DFieldInfo = Enzo1DFieldContainer()
add_enzo_1d_field = Enzo1DFieldInfo.add_field

def _CellLength(field, data):
    return data["dx"]
def _ConvertCellLengthMpc(data):
    return data.convert("mpc")
def _ConvertCellLengthCGS(data):
    return data.convert("cm")
add_enzo_1d_field("CellLengthCode", units=r"\rm{BoxArea}^2",
          function=_CellLength)
add_enzo_1d_field("CellLengthMpc", units=r"\rm{Mpc}^2",
          function=_CellLength,
          convert_function=_ConvertCellLengthMpc)
add_enzo_1d_field("CellLength", units=r"\rm{cm}^2",
          function=_CellLength,
          convert_function=_ConvertCellLengthCGS)

for a in ["Code", "Mpc", ""]:
    Enzo1DFieldInfo["CellVolume%s" % a] = \
        Enzo1DFieldInfo["CellLength%s" % a]

def _yvel(field, data):
    return na.zeros(data["x-velocity"].shape,
                    dtype='float64')
add_enzo_1d_field("z-velocity", function=_zvel)
add_enzo_1d_field("y-velocity", function=_yvel)
